from typing import Dict, Optional

from kgx.config import get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import (
    CORE_NODE_PROPERTIES,
    CORE_EDGE_PROPERTIES,
    generate_edge_key,
    current_time_in_millis,
)

log = get_logger()


def remap_node_identifier(
    graph: BaseGraph, category: str, alternative_property: str, prefix=None
) -> BaseGraph:
    """
    Remap a node's 'id' attribute with value from a node's ``alternative_property`` attribute.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    category: string
        category referring to nodes whose 'id' needs to be remapped
    alternative_property: string
        property name from which the new value is pulled from
    prefix: string
        signifies that the value for ``alternative_property`` is a list
        and the ``prefix`` indicates which value to pick from the list

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The modified graph

    """
    mapping: Dict = {}
    for nid, data in graph.nodes(data=True):
        node_data = data.copy()
        if "category" in node_data and category not in node_data["category"]:
            continue

        if alternative_property in node_data:
            alternative_values = node_data[alternative_property]
            if isinstance(alternative_values, (list, set, tuple)):
                if prefix:
                    for v in alternative_values:
                        if prefix in v:
                            # take the first occurring value that contains the given prefix
                            mapping[nid] = {"id": v}
                            break
                else:
                    # no prefix defined; pick the 1st one from list
                    mapping[nid] = {"id": next(iter(alternative_values))}
            elif isinstance(alternative_values, str):
                if prefix:
                    if alternative_values.startswith(prefix):
                        mapping[nid] = {"id": alternative_values}
                else:
                    # no prefix defined
                    mapping[nid] = {"id": alternative_values}
            else:
                log.error(
                    f"Cannot use {alternative_values} from alternative_property {alternative_property}"
                )

    graph.set_node_attributes(graph, attributes=mapping)
    graph.relabel_nodes(graph, {k: list(v.values())[0] for k, v in mapping.items()})

    # update 'subject' of all outgoing edges
    update_edge_keys = {}
    updated_subject_values = {}
    updated_object_values = {}
    for u, v, k, edge_data in graph.edges(data=True, keys=True):
        if u is not edge_data["subject"]:
            updated_subject_values[(u, v, k)] = {"subject": u}
            update_edge_keys[(u, v, k)] = {
                "edge_key": generate_edge_key(u, edge_data["predicate"], v)
            }
        if v is not edge_data["object"]:
            updated_object_values[(u, v, k)] = {"object": v}
            update_edge_keys[(u, v, k)] = {
                "edge_key": generate_edge_key(u, edge_data["predicate"], v)
            }

    graph.set_edge_attributes(graph, attributes=updated_subject_values)
    graph.set_edge_attributes(graph, attributes=updated_object_values)
    graph.set_edge_attributes(graph, attributes=update_edge_keys)
    return graph


def remap_node_property(
    graph: BaseGraph, category: str, old_property: str, new_property: str
) -> None:
    """
    Remap the value in node ``old_property`` attribute with value
    from node ``new_property`` attribute.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    category: string
        Category referring to nodes whose property needs to be remapped
    old_property: string
        old property name whose value needs to be replaced
    new_property: string
        new property name from which the value is pulled from

    """
    mapping = {}
    if old_property in CORE_NODE_PROPERTIES:
        raise AttributeError(
            f"node property {old_property} cannot be modified as it is a core property."
        )

    for nid, data in graph.nodes(data=True):
        node_data = data.copy()
        if category in node_data and category not in node_data["category"]:
            continue
        if new_property in node_data:
            mapping[nid] = {old_property: node_data[new_property]}
    graph.set_node_attributes(graph, attributes=mapping)


def remap_edge_property(
    graph: BaseGraph, edge_predicate: str, old_property: str, new_property: str
) -> None:
    """
    Remap the value in an edge ``old_property`` attribute with value
    from edge ``new_property`` attribute.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    edge_predicate: string
        edge_predicate referring to edges whose property needs to be remapped
    old_property: string
        Old property name whose value needs to be replaced
    new_property: string
        New property name from which the value is pulled from

    """
    mapping = {}
    if old_property in CORE_EDGE_PROPERTIES:
        raise AttributeError(
            f"edge property {old_property} cannot be modified as it is a core property."
        )
    for u, v, k, data in graph.edges(data=True, keys=True):
        edge_data = data.copy()
        if edge_predicate is not edge_data["predicate"]:
            continue
        if new_property in edge_data:
            mapping[(u, v, k)] = {old_property: edge_data[new_property]}
    graph.set_edge_attributes(graph, attributes=mapping)


def fold_predicate(
    graph: BaseGraph, predicate: str, remove_prefix: bool = False
) -> None:
    """
    Fold predicate as node property where every edge with ``predicate``
    will be folded as a node property.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    predicate: str
        The predicate to fold
    remove_prefix: bool
        Whether or not to remove prefix from the predicate (``False``, by default)

    """
    node_cache = []
    edge_cache = []
    start = current_time_in_millis()
    p = predicate.split(":", 1)[1] if remove_prefix else predicate
    for u, v, k, data in graph.edges(keys=True, data=True):
        if data["predicate"] == predicate:
            node_cache.append((u, p, v))
            edge_cache.append((u, v, k))
    while node_cache:
        n = node_cache.pop()
        graph.add_node_attribute(*n)
    while edge_cache:
        e = edge_cache.pop()
        graph.remove_edge(*e)
    end = current_time_in_millis()
    log.info(f"Time taken: {end - start} ms")


def unfold_node_property(
    graph: BaseGraph, node_property: str, prefix: Optional[str] = None
) -> None:
    """
    Unfold node property as a predicate where every node with ``node_property``
    will be unfolded as an edge.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    node_property: str
        The node property to unfold
    prefix: Optional[str]
        The prefix to use

    """
    node_cache = []
    edge_cache = []
    start = current_time_in_millis()
    p = f"{prefix}:{node_property}" if prefix else node_property
    for n, data in graph.nodes(data=True):
        sub = n
        if node_property in data:
            obj = data[node_property]
            edge_cache.append((sub, obj, p))
            node_cache.append((n, node_property))
    while edge_cache:
        e = edge_cache.pop()
        graph.add_edge(
            *e, **{"subject": e[0], "object": e[1], "predicate": e[2], "relation": e[2]}
        )
    while node_cache:
        n = node_cache.pop()
        del graph.nodes()[n[0]][n[1]]
    end = current_time_in_millis()
    log.info(f"Time taken: {end - start} ms")


def remove_singleton_nodes(graph: BaseGraph) -> None:
    """
    Remove singleton nodes (nodes that have a degree of 0) from the graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph

    """
    start = current_time_in_millis()
    singleton = []
    for n, d in graph.degree():
        if d == 0:
            singleton.append(n)
    while singleton:
        n = singleton.pop()
        log.debug(f"Removing singleton node {n}")
        graph.remove_node(n)
    end = current_time_in_millis()
    log.info(f"Time taken: {end - start} ms")



from typing import List, Union, Any

from neo4jrestclient.client import GraphDatabase
from neo4jrestclient.query import CypherException

from kgx.config import get_logger
from kgx.sink.sink import Sink
from kgx.utils.kgx_utils import DEFAULT_NODE_CATEGORY

log = get_logger()


class NeoSink(Sink):
    """
    NeoSink is responsible for writing data as records
    to a Neo4j instance.

    Parameters
    ----------
    uri: str
        The URI for the Neo4j instance.
        For example, http://localhost:7474
    username: str
        The username
    password: str
        The password
    kwargs: Any
        Any additional arguments

    """

    CACHE_SIZE = 100000
    node_cache = {}
    edge_cache = {}
    node_count = 0
    edge_count = 0
    CATEGORY_DELIMITER = "|"
    CYPHER_CATEGORY_DELIMITER = ":"
    _seen_categories = set()

    def __init__(self, uri: str, username: str, password: str, **kwargs: Any):
        super().__init__()
        if "cache_size" in kwargs:
            self.CACHE_SIZE = kwargs["cache_size"]
        self.http_driver: GraphDatabase = GraphDatabase(
            uri, username=username, password=password
        )

    def _flush_node_cache(self):
        self._write_node_cache()
        self.node_cache.clear()
        self.node_count = 0

    def write_node(self, record) -> None:
        """
        Cache a node record that is to be written to Neo4j.
        This method writes a cache of node records when the
        total number of records exceeds ``CACHE_SIZE``

        Parameters
        ----------
        record: Dict
            A node record

        """
        sanitized_category = self.sanitize_category(record["category"])
        category = self.CATEGORY_DELIMITER.join(sanitized_category)
        if self.node_count >= self.CACHE_SIZE:
            self._flush_node_cache()
        if category not in self.node_cache:
            self.node_cache[category] = [record]
        else:
            self.node_cache[category].append(record)
        self.node_count += 1

    def _write_node_cache(self) -> None:
        """
        Write cached node records to Neo4j.
        """
        batch_size = 10000
        categories = self.node_cache.keys()
        filtered_categories = [x for x in categories if x not in self._seen_categories]
        self.create_constraints(filtered_categories)
        for category in self.node_cache.keys():
            log.debug("Generating UNWIND for category: {}".format(category))
            cypher_category = category.replace(
                self.CATEGORY_DELIMITER, self.CYPHER_CATEGORY_DELIMITER
            )
            query = self.generate_unwind_node_query(cypher_category)
            log.debug(query)
            nodes = self.node_cache[category]
            for x in range(0, len(nodes), batch_size):
                y = min(x + batch_size, len(nodes))
                log.debug(f"Batch {x} - {y}")
                batch = nodes[x:y]
                try:
                    self.http_driver.query(query, params={"nodes": batch})
                except CypherException as ce:
                    log.error(ce)

    def _flush_edge_cache(self):
        self._flush_node_cache()
        self._write_edge_cache()
        self.edge_cache.clear()
        self.edge_count = 0

    def write_edge(self, record) -> None:
        """
        Cache an edge record that is to be written to Neo4j.
        This method writes a cache of edge records when the
        total number of records exceeds ``CACHE_SIZE``

        Parameters
        ----------
        record: Dict
            An edge record

        """
        if self.edge_count >= self.CACHE_SIZE:
            self._flush_edge_cache()
        # self.validate_edge(data)
        edge_predicate = record["predicate"]
        if edge_predicate in self.edge_cache:
            self.edge_cache[edge_predicate].append(record)
        else:
            self.edge_cache[edge_predicate] = [record]
        self.edge_count += 1

    def _write_edge_cache(self) -> None:
        """
        Write cached edge records to Neo4j.
        """
        batch_size = 10000
        for predicate in self.edge_cache.keys():
            query = self.generate_unwind_edge_query(predicate)
            log.debug(query)
            edges = self.edge_cache[predicate]
            for x in range(0, len(edges), batch_size):
                y = min(x + batch_size, len(edges))
                batch = edges[x:y]
                log.debug(f"Batch {x} - {y}")
                try:
                    self.http_driver.query(
                        query, params={"relationship": predicate, "edges": batch}
                    )
                except CypherException as ce:
                    log.error(ce)

    def finalize(self) -> None:
        """
        Write any remaining cached node and/or edge records.
        """
        self._write_node_cache()
        self._write_edge_cache()

    @staticmethod
    def sanitize_category(category: List) -> List:
        """
        Sanitize category for use in UNWIND cypher clause.
        This method adds escape characters to each element in category
        list to ensure the category is processed correctly.

        Parameters
        ----------
        category: List
            Category

        Returns
        -------
        List
            Sanitized category list

        """
        return [f"`{x}`" for x in category]

    @staticmethod
    def generate_unwind_node_query(category: str) -> str:
        """
        Generate UNWIND cypher query for saving nodes into Neo4j.

        There should be a CONSTRAINT in Neo4j for ``self.DEFAULT_NODE_CATEGORY``.
        The query uses ``self.DEFAULT_NODE_CATEGORY`` as the node label to increase speed for adding nodes.
        The query also sets label to ``self.DEFAULT_NODE_CATEGORY`` for any node to make sure that the CONSTRAINT applies.

        Parameters
        ----------
        category: str
            Node category

        Returns
        -------
        str
            The UNWIND cypher query

        """
        query = f"""
        UNWIND $nodes AS node
        MERGE (n:`{DEFAULT_NODE_CATEGORY}` {{id: node.id}})
        ON CREATE SET n += node, n:{category}
        ON MATCH SET n += node, n:{category}
        """

        return query

    @staticmethod
    def generate_unwind_edge_query(edge_predicate: str) -> str:
        """
        Generate UNWIND cypher query for saving edges into Neo4j.

        Query uses ``self.DEFAULT_NODE_CATEGORY`` to quickly lookup the required subject and object node.

        Parameters
        ----------
        edge_predicate: str
            Edge label as string

        Returns
        -------
        str
            The UNWIND cypher query

        """

        query = f"""
        UNWIND $edges AS edge
        MATCH (s:`{DEFAULT_NODE_CATEGORY}` {{id: edge.subject}}), (o:`{DEFAULT_NODE_CATEGORY}` {{id: edge.object}})
        MERGE (s)-[r:`{edge_predicate}`]->(o)
        SET r += edge
        """
        return query

    def create_constraints(self, categories: Union[set, list]) -> None:
        """
        Create a unique constraint on node 'id' for all ``categories`` in Neo4j.

        Parameters
        ----------
        categories: Union[set, list]
            Set of categories

        """
        categories_set = set(categories)
        categories_set.add(f"`{DEFAULT_NODE_CATEGORY}`")
        for category in categories_set:
            if self.CATEGORY_DELIMITER in category:
                subcategories = category.split(self.CATEGORY_DELIMITER)
                self.create_constraints(subcategories)
            else:
                query = NeoSink.create_constraint_query(category)
                try:
                    self.http_driver.query(query)
                    self._seen_categories.add(category)
                except CypherException as ce:
                    log.error(ce)

    @staticmethod
    def create_constraint_query(category: str) -> str:
        """
        Create a Cypher CONSTRAINT query

        Parameters
        ----------
        category: str
            The category to create a constraint on

        Returns
        -------
        str
            The Cypher CONSTRAINT query

        """
        query = f"CREATE CONSTRAINT ON (n:{category}) ASSERT n.id IS UNIQUE"
        return query



import gzip
import re
import jsonlines
from typing import Optional, Any, Generator, Dict

from kgx.config import get_logger

log = get_logger()

from kgx.source.json_source import JsonSource


class JsonlSource(JsonSource):
    """
    JsonlSource is responsible for reading data as records
    from JSON Lines.
    """

    def __init__(self):
        super().__init__()

    def parse(
        self,
        filename: str,
        format: str = "jsonl",
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from JSON Lines and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``json``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """

        self.set_provenance_map(kwargs)

        if re.search(f"nodes.{format}", filename):
            m = self.read_node
        elif re.search(f"edges.{format}", filename):
            m = self.read_edge
        else:
            # This used to throw an exception but perhaps we should simply ignore it.
            log.warning(
                f"Parse function cannot resolve the KGX file type in name {filename}. Skipped..."
            )
            return

        if compression == "gz":
            with gzip.open(filename, "rb") as FH:
                reader = jsonlines.Reader(FH)
                for obj in reader:
                    yield m(obj)
        else:
            with jsonlines.open(filename) as FH:
                for obj in FH:
                    yield m(obj)



from .source import Source
from .tsv_source import TsvSource
from .json_source import JsonSource
from .jsonl_source import JsonlSource
from .obograph_source import ObographSource
from .trapi_source import TrapiSource
from .neo_source import NeoSource
from .rdf_source import RdfSource
from .graph_source import GraphSource
from .owl_source import OwlSource
from .sssom_source import SssomSource



"""
Biolink 2.0 Information Resource (InfoRes) utilities
"""
import re
from typing import Optional, Tuple, Callable, Dict, List, Any

from kgx.utils.kgx_utils import knowledge_provenance_properties, column_types


class InfoResContext:
    """
    Information Resource CURIE management context for knowledge sources.
    """

    def __init__(self):

        self.default_provenance = "Graph"

        # this dictionary captures the operational mappings
        # for a specified knowledge source field in the graph
        self.mapping: Dict[str, Any] = dict()

        # this dictionary records specific knowledge source
        # name to infores associations for the given graph
        self.catalog: Dict[str, str] = dict()

    def get_catalog(self) -> Dict[str, str]:
        """
        Retrieves the catalog of mappings of Knowledge Source names to an InfoRes.

        Returns
        -------
        Dict[str, str]
            Dictionary where the index string is Knowledge Source Names and values are the corresponding InfoRes CURIE

        """
        return self.catalog

    class InfoResMapping:
        """
        Knowledge Source mapping onto an Information Resource identifier.
        """

        def __init__(self, context, ksf: str):
            """
            InfoRes mapping specification for a single knowledge_source (or related) field

            Parameters
            ----------
            context: InfoResContext
                The KGX knowledge graph and default configuration context within which this InfoResMapping exists.
            ksf: str
                Knowledge Source Field being processed.

            """
            self.context = context  # parent InfoRes context
            self.ksf = ksf  # Biolink 2.* 'Knowledge Source Field' slot name
            self.filter = None
            self.substr = ""
            self.prefix = ""

        def processor(self, infores_rewrite_filter: Optional[Tuple] = None) -> Callable:
            """
            Full processor of a Knowledge Source name into an InfoRes. The conversion is made based on
            client-caller specified rewrite rules for a given knowledge source field ('ksf').

            Parameters
            ----------
            infores_rewrite_filter: Optional[Tuple]
                The presence of this optional Tuple argument signals an InfoRes rewrite of any
                Biolink 2.0 compliant knowledge source field name in node and edge data records.
                The mere presence of a (possibly empty) Tuple signals a rewrite. If the Tuple is empty,
                then only a standard transformation of the field value is performed. If the Tuple has
                an infores_rewrite[0] value, it is assumed to be a regular expression (string) to match
                against. If there is no infores_rewrite[1] value or it is empty, then matches of the
                infores_rewrite[0] are simply deleted from the field value prior to coercing the field
                value into an InfoRes CURIE. Otherwise, a non-empty second string value of infores_rewrite[1]
                is a substitution string for the regex value matched in the field. If the Tuple contains
                a third non-empty string (as infores_rewrite[2]), then the given string is added as a prefix
                to the InfoRes.  Whatever the transformations, unique InfoRes identifiers once generated,
                are used in the meta_knowledge_graph and also reported using the get_infores_catalog() method.

            Returns
            -------
            Callable
                A locally configured Callable that knows how to process
                a source name string into an infores CURIE, using on client-specified
                rewrite rules applied alongside standard formatting rules.

            """
            # Check for non-empty infores_rewrite_filter
            if infores_rewrite_filter:
                self.filter = (
                    re.compile(infores_rewrite_filter[0])
                    if infores_rewrite_filter[0]
                    else None
                )
                self.substr = (
                    infores_rewrite_filter[1] if len(infores_rewrite_filter) > 1 else ""
                )
                self.prefix = (
                    infores_rewrite_filter[2] if len(infores_rewrite_filter) > 2 else ""
                )

            def _get_infores(source: str) -> str:
                """
                Get InfoRes CURIE inferred from source name.

                Parameters
                ----------
                source: str
                    Name of Information Resource associated with the InfoRes
                    (i.e. from which the InfoRes was inferred)

                Returns
                -------
                str:
                    infores CURIE, retrieved or generated.

                """
                if source in self.context.catalog:
                    return self.context.catalog[source]
                else:
                    infores: str = _process_infores(source)
                    if infores:
                        self.context.catalog[source] = infores
                        return infores
                    else:
                        return ""

            def _process_infores(source: str) -> str:
                """
                Process a single knowledge Source name  string into an infores, by applying rules
                in the _infores_processor() closure context, followed by standard formatting.

                Parameters
                ----------
                source: str
                    Knowledge source name string being processed.

                Returns
                -------
                str
                    Infores CURIE inferred from the input Knowledge Source name string.

                """
                # don't touch something that already looks like an infores CURIE
                if source.startswith("infores:"):
                    return source

                if self.filter:
                    infores = self.filter.sub(self.substr, source)
                else:
                    infores = source
                infores = self.prefix + " " + infores
                infores = infores.strip()
                infores = infores.lower()
                infores = re.sub(r"\s+", "_", infores)
                infores = re.sub(r"\.+", "_", infores)
                infores = re.sub(r"[\W]", "", infores)
                infores = re.sub(r"_", "-", infores)

                # TODO: to be fully compliant, the InfoRes needs to have the 'infores' prefix?
                infores = "infores:" + infores

                return infores

            def parser_list(sources: Optional[List[str]] = None) -> List[str]:
                """
                Infores parser for a list of input knowledge source names.

                Parameters
                ----------
                sources: List[str]
                    List of Knowledge source name strings being processed.

                Returns
                -------
                List[str]
                    Source name strings transformed into infores CURIES, using _process_infores().

                """
                if not sources:
                    return [self.context.default_provenance]
                results: List[str] = list()
                for source in sources:
                    infores = _get_infores(source)
                    if infores:
                        results.append(infores)
                return results

            def parser_scalar(source=None) -> str:
                """
                Infores parser for a single knowledge source name string.

                Parameters
                ----------
                source: str
                    Knowledge source name string being processed.

                Returns
                -------
                str
                    Source name string transformed into an infores CURIE, using _process_infores().

                """
                return (
                    self.context.default_provenance
                    if not source
                    else _get_infores(source)
                )

            if self.ksf in column_types and column_types[self.ksf] == list:
                return parser_list
            else:
                # not sure how safe an assumption for all non-list column_types, but...
                return parser_scalar

        def default(self, default=None):
            """
            Lightweight alternative to the KS processor() which simply assigns knowledge_source fields
            simple client-defined default knowledge source strings (not constrained to be formatted as infores CURIEs).

            Parameters
            ----------
            default: str
                (Optional) default value of the knowledge source field.

            Returns
            -------
            Callable
                A locally configured Callable that knows how to process a source name string
                (possibly empty) into a suitable (possibly default) infores string identifier.

            """

            def default_value_list(sources: List[str] = None):
                """
                Infores default method for a list of input knowledge source names.

                Parameters
                ----------
                sources: List[str]
                    List of Knowledge source name strings being processed.

                Returns
                -------
                List[str]
                    Infores identifiers mapped to input source strings.

                """
                if not default:
                    return list()
                if not sources:
                    return [default]
                else:
                    return sources

            def default_value_scalar(source=None):
                """
                Infores default method for single input knowledge source name.

                Parameters
                ----------
                source: str
                    Knowledge source name string being processed.

                Returns
                -------
                str
                    Infores identifier mapped to the input source string.

                """
                if not default:
                    return None
                if not source:
                    return default
                else:
                    return source

            if self.ksf in column_types and column_types[self.ksf] == list:
                return default_value_list
            else:
                # not sure how safe an assumption for non-list column_types, but...
                return default_value_scalar

        def set_provenance_map_entry(self, ksf_value: Any) -> Any:
            """
            Set up a provenance (Knowledge Source to InfoRes) map entry
            """
            if isinstance(ksf_value, str):
                ksf_value = ksf_value.strip()
                if ksf_value.lower() == "true":
                    mapping = self.processor()
                elif ksf_value.lower() == "false":
                    mapping = self.default()  # source suppressed
                else:
                    mapping = self.default(ksf_value)
            elif isinstance(ksf_value, bool):
                if ksf_value:
                    mapping = self.processor()
                else:  # false, ignore this source?
                    mapping = self.default()  # source suppressed
            elif isinstance(ksf_value, (list, set, tuple)):
                mapping = self.processor(infores_rewrite_filter=ksf_value)
            else:
                # Not sure what to do here... just return the original ksf_value?
                mapping = ksf_value
            return mapping

    def get_mapping(self, ksf: str) -> InfoResMapping:
        """
        InfoRes mapping for a specified knolwedge source field ('ksf').

        Parameters
        ----------
        ksf: str
            Knowledge Source Field whose mapping is being managed.

        """
        irm = self.InfoResMapping(self, ksf)
        return irm

    def set_provenance_map(self, kwargs: Dict):
        """
        A knowledge_source property indexed map set up with various mapping
        Callable methods to process input knowledge source values into
        suitable InfoRes identifiers.

        Parameters
        ----------
        kwargs: Dict
            The input keyword argument dictionary was likely propagated from the
            Transformer.transform() method input_args, and is here harvested for
            static defaults or rewrite rules for knowledge_source slot InfoRes value processing.

        """
        if "default_provenance" in kwargs:
            self.default_provenance = kwargs.pop("default_provenance")

        # Biolink 2.0 knowledge_source 'knowledge_source' derived fields
        ksf_found = False
        for ksf in knowledge_provenance_properties:
            if ksf in kwargs:
                if not ksf_found:
                    ksf_found = ksf  # save the first one found, for later
                ksf_value = kwargs.pop(ksf)
                # Check if the ksf_value is a multi-valued catalog of patterns for a
                # given knowledge graph field, indexed on each distinct regex pattern
                if isinstance(ksf_value, dict):
                    for ksf_pattern in ksf_value.keys():
                        if ksf not in self.mapping:
                            self.mapping[ksf] = dict()
                        ir = self.get_mapping(ksf)
                        self.mapping[ksf][ksf_pattern] = ir.set_provenance_map_entry(
                            ksf_value[ksf_pattern]
                        )
                else:
                    ir = self.get_mapping(ksf)
                    self.mapping[ksf] = ir.set_provenance_map_entry(ksf_value)

        # if none specified, add at least one generic 'knowledge_source'
        if not ksf_found:
            ksf_found = "knowledge_source"  # knowledge source field 'ksf' is set, one way or another
            ir = self.get_mapping(ksf_found)
            if "name" in kwargs:
                self.mapping["knowledge_source"] = ir.default(kwargs["name"])
            else:
                self.mapping["knowledge_source"] = ir.default(self.default_provenance)

        # TODO: better to lobby the team to totally deprecated this, even for Nodes?
        if "provided_by" not in self.mapping:
            self.mapping["provided_by"] = self.mapping[ksf_found]

    def set_provenance(self, ksf: str, data: Dict):
        """
        Compute the knowledge_source value for the current node or edge data, using the
        infores rewrite context previously established by a call to set_provenance_map().

        Parameters
        ----------
        ksf: str
            Knowledge source field being processed.
        data: Dict
            Current node or edge data entry being processed.

        """
        if ksf not in data.keys():
            if ksf in self.mapping and not isinstance(self.mapping[ksf], dict):
                data[ksf] = self.mapping[ksf]()  # get default ksf value?
            else:
                # if unknown ksf or is an inapplicable pattern
                # dictionary, then just set the value to the default
                data[ksf] = [self.default_provenance]
        else:  # valid data value but... possible InfoRes rewrite?
            # If data is s a non-string iterable
            # then, coerce into a simple list of sources
            if isinstance(data[ksf], (list, set, tuple)):
                sources = list(data[ksf])
            else:
                # Otherwise, just assumed to be a scalar
                # data value for this knowledge source field?
                # Treat differentially depending on column type...
                if column_types[ksf] == list:
                    sources = [data[ksf]]
                else:
                    sources = data[ksf]
            if ksf in self.mapping:
                if isinstance(self.mapping[ksf], dict):
                    # Need to iterate through a knowledge source pattern dictionary
                    for pattern in self.mapping[ksf].keys():
                        for source in sources:
                            # TODO: I need to test pattern match to each source?
                            if re.compile(pattern).match(source):
                                data[ksf] = self.mapping[ksf][pattern]([source])
                            if data[ksf]:
                                break
                else:
                    data[ksf] = self.mapping[ksf](sources)
            else:  # leave data intact?
                data[ksf] = sources

        # ignore if still empty at this point
        if not data[ksf]:
            data.pop(ksf)

    def set_node_provenance(self, node_data: Dict):
        """
        Sets the node knowledge_source value for the current node. At the moment, nodes are still
        hard-coded to using the (Biolink 2.0 deprecated) 'provided_by' knowledge_source property;
        However, this could change in the future to use edge 'knowledge_source' properties.

        Parameters
        ----------
        node_data: Dict
            Current node data entry being processed.

        """
        self.set_provenance("provided_by", node_data)

    # TODO: need to design a more efficient algorithm here...
    def set_edge_provenance(self, edge_data: Dict):
        """
        Sets the node knowledge_source value for the current node. Edge knowledge_source properties
        include the full Biolink 2.0 'knowledge_source' related properties.

        Parameters
        ----------
        edge_data: Dict
            Current edge data entry being processed.

        """
        ksf_found = False
        data_fields = list(edge_data.keys())
        for ksf in data_fields:
            if ksf in knowledge_provenance_properties:
                ksf_found = True
                self.set_provenance(ksf, edge_data)
        if not ksf_found:
            for ksf in self.mapping:
                if ksf != "provided_by":
                    self.set_provenance(ksf, edge_data)



import importlib
from typing import Dict, Any, Optional
import sys
from os import path

import re

import requests
import yaml
import json
import logging

from kgx.graph.base_graph import BaseGraph

config: Optional[Dict[str, Any]] = None
logger: Optional[logging.Logger] = None
graph_store_class: Optional[BaseGraph] = None
jsonld_context_map: Dict = {}

CONFIG_FILENAME = path.join(path.dirname(path.abspath(__file__)), "config.yml")


def get_config(filename: str = CONFIG_FILENAME) -> Dict:
    """
    Get config as a Dictionary

    Parameters
    ----------
    filename: str
        The filename with all the configuration

    Returns
    -------
    Dict
        A Dictionary containing all the entries from the config YAML

    """
    global config
    if config is None:
        config = yaml.load(open(filename), Loader=yaml.FullLoader)
    return config


def get_jsonld_context(name: str = "biolink"):
    """
    Get contents of a JSON-LD context.

    Returns
    -------
    Dict
        the contents of the JSON-LD context

    """
    content = None
    if name in jsonld_context_map:
        content = jsonld_context_map[name]
    else:
        filepath = config["jsonld-context"][name]  # type: ignore
        if filepath.startswith("http"):
            try:
                content = requests.get(filepath).json()
            except ConnectionError:
                raise Exception(f"Unable to download JSON-LD context from {filepath}")
        else:
            if path.exists(filepath):
                content = json.load(open(filepath))

        if "@context" in content:
            content = content["@context"]
            jsonld_context_map[name] = content
    return content


def get_logger(name: str = "KGX") -> logging.Logger:
    """
    Get an instance of logger.

    Parameters
    ----------
    name: str
        The name of logger

    Returns
    -------
    logging.Logger
        An instance of logging.Logger

    """
    global logger
    if logger is None:
        config = get_config()
        logger = logging.getLogger(name)
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(config["logging"]["format"])
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(config["logging"]["level"])
        logger.propagate = False
    return logger


def get_graph_store_class() -> Any:
    """
    Get a reference to the graph store class, as defined the config.
    Defaults to ``kgx.graph.nx_graph.NxGraph``

    Returns
    -------
    Any
        A reference to the graph store class

    """
    global graph_store_class
    if not graph_store_class:
        config = get_config()
        if "graph_store" in config:
            name = config["graph_store"]
        else:
            name = "kgx.graph.nx_graph.NxGraph"
        module_name = ".".join(name.split(".")[0:-1])
        class_name = name.split(".")[-1]
        graph_store_class = getattr(importlib.import_module(module_name), class_name)
    return graph_store_class


# Biolink Release number should be a well formed Semantic Versioning (patch is optional?)
semver_pattern = re.compile(r"^\d+\.\d+\.\d+$")


def get_biolink_model_schema(biolink_release: Optional[str] = None) -> Optional[str]:
    if biolink_release:
        if not semver_pattern.fullmatch(biolink_release):
            raise TypeError(
                "The 'biolink_release' argument '"
                + biolink_release
                + "' is not a properly formatted 'major.minor.patch' semantic version?"
            )
        schema = f"https://raw.githubusercontent.com/biolink/biolink-model/{biolink_release}/biolink-model.yaml"
        return schema
    else:
        return None



from typing import Any
from kgx.sink import Sink


class NullSink(Sink):
    """
     A NullSink just ignores any date written to it,
     effectively a /dev/null device for Transformer
     data flows, in which the inspection of the input
     knowledge graph is the important operation, but
     the graph itself is not persisted in the output
     (in particular, not in memory, where the huge
     memory footprint may be problematics, e.g. when
     stream processing huge graphs).

    Parameters
    ----------
    n/a (**kwargs allowed, but ignored)
    """

    def __init__(self, **kwargs: Any):
        super().__init__()

    def write_node(self, record) -> None:
        """
        Write a node record to the underlying store.

        Parameters
        ----------
        record: Any
            A node record

        """
        pass

    def write_edge(self, record) -> None:
        """
        Write an edge record to the underlying store.

        Parameters
        ----------
        record: Any
            An edge record

        """
        pass

    def finalize(self) -> None:
        """
        Operations that ought to be done after
        writing all the incoming data should be called
        by this method.

        """
        pass



from typing import Dict, Union, Optional

from kgx.utils.infores import InfoResContext
from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger

log = get_logger()


class Source(object):
    """
    A Source is responsible for reading data as records
    from a store where the store is a file or a database.
    """

    def __init__(self):
        self.graph_metadata: Dict = {}
        self.node_filters = {}
        self.edge_filters = {}
        self.node_properties = set()
        self.edge_properties = set()
        self.prefix_manager = PrefixManager()
        self.infores_context: Optional[InfoResContext] = None

    def set_prefix_map(self, m: Dict) -> None:
        """
        Update default prefix map.

        Parameters
        ----------
        m: Dict
            A dictionary with prefix to IRI mappings

        """
        self.prefix_manager.update_prefix_map(m)

    def check_node_filter(self, node: Dict) -> bool:
        """
        Check if a node passes defined node filters.

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        bool
            Whether the given node has passed all defined node filters

        """
        pass_filter = False
        if self.node_filters:
            for k, v in self.node_filters.items():
                if k in node:
                    # filter key exists in node
                    if isinstance(v, (list, set, tuple)):
                        if any(x in node[k] for x in v):
                            pass_filter = True
                        else:
                            return False
                    elif isinstance(v, str):
                        if node[k] == v:
                            pass_filter = True
                        else:
                            return False
                    else:
                        log.error(f"Unexpected {k} node filter of type {type(v)}")
                        return False
                else:
                    # filter key does not exist in node
                    return False
        else:
            # no node filters defined
            pass_filter = True
        return pass_filter

    def check_edge_filter(self, edge: Dict) -> bool:
        """
        Check if an edge passes defined edge filters.

        Parameters
        ----------
        edge: Dict
            An edge

        Returns
        -------
        bool
            Whether the given edge has passed all defined edge filters

        """
        pass_filter = False
        if self.edge_filters:
            for k, v in self.edge_filters.items():
                if k in {"subject_category", "object_category"}:
                    pass_filter = True
                    continue
                if k in edge:
                    # filter key exists in edge
                    if isinstance(v, (list, set, tuple)):
                        if any(x in edge[k] for x in v):
                            pass_filter = True
                        else:
                            return False
                    elif isinstance(v, str):
                        if edge[k] == v:
                            pass_filter = True
                        else:
                            return False
                    else:
                        log.error(f"Unexpected {k} edge filter of type {type(v)}")
                        return False
                else:
                    # filter does not exist in edge
                    return False
        else:
            # no edge filters defined
            pass_filter = True
        return pass_filter

    def set_node_filter(self, key: str, value: Union[str, set]) -> None:
        """
        Set a node filter, as defined by a key and value pair.
        These filters are used to filter (or reduce) the
        search space when fetching nodes from the underlying store.

        .. note::
            When defining the 'category' filter, the value should be of type ``set``.
            This method also sets the 'subject_category' and 'object_category'
            edge filters, to get a consistent set of nodes in the subgraph.

        Parameters
        ----------
        key: str
            The key for node filter
        value: Union[str, set]
            The value for the node filter.
            Can be either a string or a set.

        """
        if key == "category":
            if isinstance(value, set):
                if "subject_category" in self.edge_filters:
                    self.edge_filters["subject_category"].update(value)
                else:
                    self.edge_filters["subject_category"] = value
                if "object_category" in self.edge_filters:
                    self.edge_filters["object_category"].update(value)
                else:
                    self.edge_filters["object_category"] = value
            else:
                raise TypeError(
                    "'category' node filter should have a value of type 'set'"
                )

        if key in self.node_filters:
            self.node_filters[key].update(value)
        else:
            self.node_filters[key] = value

    def set_node_filters(self, filters: Dict) -> None:
        """
        Set node filters.

        Parameters
        ----------
        filters: Dict
            Node filters

        """
        if filters:
            for k, v in filters.items():
                if isinstance(v, (list, set, tuple)):
                    self.set_node_filter(k, set(v))
                else:
                    self.set_node_filter(k, v)

    def set_edge_filters(self, filters: Dict) -> None:
        """
        Set edge filters.

        Parameters
        ----------
        filters: Dict
            Edge filters

        """
        if filters:
            for k, v in filters.items():
                if isinstance(v, (list, set, tuple)):
                    self.set_edge_filter(k, set(v))
                else:
                    self.set_edge_filter(k, v)

    def set_edge_filter(self, key: str, value: set) -> None:
        """
        Set an edge filter, as defined by a key and value pair.
        These filters are used to filter (or reduce) the
        search space when fetching nodes from the underlying store.

        .. note::
            When defining the 'subject_category' or 'object_category' filter,
            the value should be of type ``set``.
            This method also sets the 'category' node filter, to get a
            consistent set of nodes in the subgraph.

        Parameters
        ----------
        key: str
            The key for edge filter
        value: Union[str, set]
            The value for the edge filter.
            Can be either a string or a set.

        """
        if key in {"subject_category", "object_category"}:
            if isinstance(value, set):
                if "category" in self.node_filters:
                    self.node_filters["category"].update(value)
                else:
                    self.node_filters["category"] = value
            else:
                raise TypeError(
                    f"'{key}' edge filter should have a value of type 'set'"
                )

        if key in self.edge_filters:
            self.edge_filters[key].update(value)
        else:
            self.edge_filters[key] = value

    def clear_graph_metadata(self):
        """
        Clears a Source graph's internal graph_metadata. The value of such graph metadata is (now)
        generally a Callable function. This operation can be used in the code when the metadata is
        no longer needed, but may cause peculiar Python object persistent problems downstream.
        """
        self.infores_context = None

    def set_provenance_map(self, kwargs):
        """
        Set up a provenance (Knowledge Source to InfoRes) map
        """
        self.infores_context = InfoResContext()
        self.infores_context.set_provenance_map(kwargs)

    def get_infores_catalog(self) -> Dict[str, str]:
        """
        Return the InfoRes Context of the source
        """
        if not self.infores_context:
            return dict()
        return self.infores_context.get_catalog()

    def set_node_provenance(self, node_data):
        """
        Set a specific node provenance value.
        """
        self.infores_context.set_node_provenance(node_data)

    def set_edge_provenance(self, edge_data):
        """
        Set a specific edge provenance value.
        """
        self.infores_context.set_edge_provenance(edge_data)



from collections import OrderedDict
from typing import List, Optional, Any, Union, Dict, Tuple
import rdflib
from linkml_runtime.linkml_model.meta import Element, SlotDefinition, ClassDefinition
from cachetools import cached, LRUCache
from rdflib import Namespace, URIRef
from rdflib.namespace import RDF, RDFS, OWL, SKOS

from kgx.config import get_logger
from kgx.prefix_manager import PrefixManager
from kgx.utils.graph_utils import get_category_via_superclass
from kgx.utils.kgx_utils import (
    get_curie_lookup_service,
    contract,
    get_cache,
    get_toolkit,
    sentencecase_to_snakecase,
    sentencecase_to_camelcase,
    get_biolink_ancestors,
)

log = get_logger()

OBAN = Namespace("http://purl.org/oban/")
BIOLINK = Namespace("https://w3id.org/biolink/vocab/")
OIO = Namespace("http://www.geneontology.org/formats/oboInOwl#")
OBO = Namespace("http://purl.obolibrary.org/obo/")

property_mapping: Dict = dict()
reverse_property_mapping: Dict = dict()

# TODO: this should be populated via bmt
is_property_multivalued = {
    "id": False,
    "subject": False,
    "object": False,
    "predicate": False,
    "description": False,
    "synonym": True,
    "in_taxon": False,
    "same_as": True,
    "name": False,
    "has_evidence": False,
    "provided_by": True,
    "category": True,
    "publications": True,
    "type": False,
    "relation": False,
}


top_level_terms = {
    OBO.term("CL_0000000"): "cell",
    OBO.term("UBERON_0001062"): "anatomical_entity",
    OBO.term("PATO_0000001"): "quality",
    OBO.term("NCBITaxon_131567"): "organism",
    OBO.term("CLO_0000031"): "cell_line",
    OBO.term("MONDO_0000001"): "disease",
    OBO.term("CHEBI_23367"): "molecular_entity",
    OBO.term("CHEBI_23888"): "drug",
    OBO.term("UPHENO_0001001"): "phenotypic_feature",
    OBO.term("GO_0008150"): "biological_process",
    OBO.term("GO_0009987"): "cellular_process",
    OBO.term("GO_0005575"): "cellular_component",
    OBO.term("GO_0003674"): "molecular_function",
    OBO.term("SO_0000704"): "gene",
    OBO.term("GENO_0000002"): "variant_locus",
    OBO.term("GENO_0000536"): "genotype",
    OBO.term("SO_0000110"): "sequence_feature",
    OBO.term("ECO_0000000"): "evidence",
    OBO.term("PW_0000001"): "pathway",
    OBO.term("IAO_0000310"): "publication",
    OBO.term("SO_0001483"): "snv",
    OBO.term("GENO_0000871"): "haplotype",
    OBO.term("SO_0001024"): "haplotype",
    OBO.term("SO_0000340"): "chromosome",
    OBO.term("SO_0000104"): "protein",
    OBO.term("SO_0001500"): "phenotypic_marker",
    OBO.term("SO_0000001"): "region",
    OBO.term("HP_0032223"): "blood_group",
    OBO.term("HP_0031797"): "clinical_course",
    OBO.term("HP_0040279"): "frequency",
    OBO.term("HP_0000118"): "phenotypic_abnormality",
    OBO.term("HP_0032443"): "past_medical_history",
    OBO.term("HP_0000005"): "mode_of_inheritance",
    OBO.term("HP_0012823"): "clinical_modifier",
}


def infer_category(iri: URIRef, rdfgraph: rdflib.Graph) -> Optional[List]:
    """
    Infer category for a given iri by traversing rdfgraph.

    Parameters
    ----------
    iri: rdflib.term.URIRef
        IRI
    rdfgraph: rdflib.Graph
        A graph to traverse

    Returns
    -------
    Optional[List]
        A list of category corresponding to the given IRI

    """
    closure = list(rdfgraph.transitive_objects(iri, URIRef(RDFS.subClassOf)))
    category = [top_level_terms[x] for x in closure if x in top_level_terms.keys()]
    if category:
        log.debug(
            "Inferred category as {} based on transitive closure over 'subClassOf' relation".format(
                category
            )
        )
    else:
        subj = closure[-1]
        if subj == iri:
            return category
        subject_curie: Optional[str] = contract(subj)
        if subject_curie and "_" in subject_curie:
            fixed_curie = subject_curie.split(":", 1)[1].split("_", 1)[1]
            log.warning(
                "Malformed CURIE {} will be fixed to {}".format(
                    subject_curie, fixed_curie
                )
            )
            subject_curie = fixed_curie
        cls = get_curie_lookup_service()
        category = list(get_category_via_superclass(cls.ontology_graph, subject_curie))
    return category


@cached(LRUCache(maxsize=1024))
def get_biolink_element(
    prefix_manager: PrefixManager, predicate: Any
) -> Optional[Element]:
    """
    Returns a Biolink Model element for a given predicate.

    Parameters
    ----------
    prefix_manager: PrefixManager
        An instance of prefix manager
    predicate: Any
        The CURIE of a predicate

    Returns
    -------
    Optional[Element]
        The corresponding Biolink Model element

    """
    toolkit = get_toolkit()
    if prefix_manager.is_iri(predicate):
        predicate_curie = prefix_manager.contract(predicate)
    else:
        predicate_curie = predicate
    if prefix_manager.is_curie(predicate_curie):
        reference = prefix_manager.get_reference(predicate_curie)
    else:
        reference = predicate_curie
    element = toolkit.get_element(reference)
    if not element:
        try:
            mapping = toolkit.get_element_by_mapping(predicate)
            if mapping:
                element = toolkit.get_element(mapping)
        except ValueError as e:
            log.error(e)
    return element


def process_predicate(
    prefix_manager: PrefixManager,
    p: Union[URIRef, str],
    predicate_mapping: Optional[Dict] = None,
) -> Tuple:
    """
    Process a predicate where the method checks if there is a mapping in Biolink Model.

    Parameters
    ----------
    prefix_manager: PrefixManager
        An instance of prefix manager
    p: Union[URIRef, str]
        The predicate
    predicate_mapping: Optional[Dict]
        Predicate mappings

    Returns
    -------
    Tuple[str, str, str, str]
        A tuple that contains the Biolink CURIE (if available), the Biolink slot_uri CURIE (if available),
        the CURIE form of p, the reference of p

    """
    if prefix_manager.is_iri(p):
        predicate = prefix_manager.contract(str(p))
    else:
        predicate = None
    if prefix_manager.is_curie(p):
        property_name = prefix_manager.get_reference(p)
        predicate = p
    else:
        if predicate and prefix_manager.is_curie(predicate):
            property_name = prefix_manager.get_reference(predicate)
        else:
            property_name = p
            predicate = f":{p}"
    element = get_biolink_element(prefix_manager, p)
    canonical_uri = None
    if element:
        if isinstance(element, SlotDefinition):
            # predicate corresponds to a biolink slot
            if element.definition_uri:
                element_uri = prefix_manager.contract(element.definition_uri)
            else:
                element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
            if element.slot_uri:
                canonical_uri = element.slot_uri
        elif isinstance(element, ClassDefinition):
            # this will happen only when the IRI is actually
            # a reference to a class
            element_uri = prefix_manager.contract(element.class_uri)
        else:
            element_uri = f"biolink:{sentencecase_to_camelcase(element.name)}"
        if "biolink:Attribute" in get_biolink_ancestors(element.name):
            element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
        if not predicate:
            predicate = element_uri
    else:
        # no mapping to biolink model;
        # look at predicate mappings
        element_uri = None
        if predicate_mapping:
            if p in predicate_mapping:
                property_name = predicate_mapping[p]
                predicate = f":{property_name}"
        # cache[p] = {'element_uri': element_uri, 'canonical_uri': canonical_uri,
        # 'predicate': predicate, 'property_name': property_name}
    return element_uri, canonical_uri, predicate, property_name



import networkx as nx
import rdflib
from kgx.config import get_logger, get_config
from kgx.utils.kgx_utils import generate_edge_key, contract

CURIE_MAP = {"BFO:0000054": "realized_in", "RO:0000091": "has_disposition"}

log = get_logger()


class CurieLookupService(object):
    """
    A service to lookup label for a given CURIE.
    """

    config = get_config()
    ontologies = config['ontologies'] if 'ontologies' in config else {}
    ontology_graph = None

    def __init__(self, curie_map: dict = None):
        if curie_map:
            self.curie_map = CURIE_MAP
            self.curie_map.update(curie_map)
        else:
            self.curie_map = CURIE_MAP
        self.ontology_graph = nx.MultiDiGraph()
        self.load_ontologies()

    def load_ontologies(self):
        """
        Load all required ontologies.
        """
        for ontology in self.ontologies.values():
            rdfgraph = rdflib.Graph()
            input_format = rdflib.util.guess_format(ontology)
            rdfgraph.parse(ontology, format=input_format)
            # triples = rdfgraph.triples((None, rdflib.RDFS.subClassOf, None))
            # for s,p,o in triples:
            #     subject_curie = contract(s)
            #     object_curie = contract(o)
            #     self.ontology_graph.add_node(subject_curie)
            #     self.ontology_graph.add_node(object_curie)
            #     key = generate_edge_key(subject_curie, 'subclass_of', object_curie)
            #     self.ontology_graph.add_edge(subject_curie, object_curie, key, **{'predicate': 'subclass_of', 'relation': 'rdfs:subClassOf'})

            triples = rdfgraph.triples((None, rdflib.RDFS.label, None))
            for s, p, o in triples:
                key = contract(s)
                value = o.value
                value = value.replace(" ", "_")
                self.curie_map[key] = value
                self.ontology_graph.add_node(key, name=value)



from typing import Dict

from kgx.prefix_manager import PrefixManager


class Sink(object):
    """
    A Sink is responsible for writing data as records
    to a store where the store is a file or a database.
    """

    def __init__(self):
        self.prefix_manager = PrefixManager()
        self.node_properties = set()
        self.edge_properties = set()

    def set_reverse_prefix_map(self, m: Dict) -> None:
        """
        Update default reverse prefix map.

        Parameters
        ----------
        m: Dict
            A dictionary with IRI to prefix mappings

        """
        self.prefix_manager.update_reverse_prefix_map(m)

    def write_node(self, record) -> None:
        """
        Write a node record to the underlying store.

        Parameters
        ----------
        record: Any
            A node record

        """
        pass

    def write_edge(self, record) -> None:
        """
        Write an edge record to the underlying store.

        Parameters
        ----------
        record: Any
            An edge record

        """
        pass

    def finalize(self) -> None:
        """
        Operations that ought to be done after
        writing all the incoming data should be called
        by this method.

        """
        pass



import os
import tarfile
from typing import Optional, Dict, Set, Any, List
from ordered_set import OrderedSet

from kgx.sink.sink import Sink
from kgx.utils.kgx_utils import (
    extension_types,
    archive_write_mode,
    archive_format,
    remove_null,
    _sanitize_export,
)


DEFAULT_NODE_COLUMNS = {"id", "name", "category", "description", "provided_by"}
DEFAULT_EDGE_COLUMNS = {
    "id",
    "subject",
    "predicate",
    "object",
    "relation",
    "category",
    "provided_by",
}


class TsvSink(Sink):
    """
    TsvSink is responsible for writing data as records to a TSV/CSV.

    Parameters
    ----------
    filename: str
        The filename to write to
    format: str
        The file format (``tsv``, ``csv``)
    compression: str
        The compression type (``tar``, ``tar.gz``)
    kwargs: Any
        Any additional arguments
    """

    def __init__(
        self,
        filename: str,
        format: str,
        compression: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()
        if format not in extension_types:
            raise Exception(f"Unsupported format: {format}")
        self.delimiter = extension_types[format]
        self.dirname = os.path.abspath(os.path.dirname(filename))
        self.basename = os.path.basename(filename)
        self.extension = format.split(":")[0]
        self.mode = (
            archive_write_mode[compression]
            if compression in archive_write_mode
            else None
        )
        self.nodes_file_basename = f"{self.basename}_nodes.{self.extension}"
        self.edges_file_basename = f"{self.basename}_edges.{self.extension}"
        if self.dirname:
            os.makedirs(self.dirname, exist_ok=True)
        if "node_properties" in kwargs:
            self.node_properties.update(set(kwargs["node_properties"]))
        else:
            self.node_properties.update(DEFAULT_NODE_COLUMNS)
        if "edge_properties" in kwargs:
            self.edge_properties.update(set(kwargs["edge_properties"]))
        else:
            self.edge_properties.update(DEFAULT_EDGE_COLUMNS)
        self.ordered_node_columns = TsvSink._order_node_columns(self.node_properties)
        self.ordered_edge_columns = TsvSink._order_edge_columns(self.edge_properties)

        self.nodes_file_name = os.path.join(
            self.dirname if self.dirname else "", self.nodes_file_basename
        )
        self.NFH = open(self.nodes_file_name, "w")
        self.NFH.write(self.delimiter.join(self.ordered_node_columns) + "\n")
        self.edges_file_name = os.path.join(
            self.dirname if self.dirname else "", self.edges_file_basename
        )
        self.EFH = open(self.edges_file_name, "w")
        self.EFH.write(self.delimiter.join(self.ordered_edge_columns) + "\n")

    def write_node(self, record: Dict) -> None:
        """
        Write a node record to the underlying store.

        Parameters
        ----------
        record: Dict
            A node record

        """
        row = self._build_export_row(record)
        row["id"] = record["id"]
        values = []
        for c in self.ordered_node_columns:
            if c in row:
                values.append(str(row[c]))
            else:
                values.append("")
        self.NFH.write(self.delimiter.join(values) + "\n")

    def write_edge(self, record: Dict) -> None:
        """
        Write an edge record to the underlying store.

        Parameters
        ----------
        record: Dict
            An edge record

        """
        row = self._build_export_row(record)
        values = []
        for c in self.ordered_edge_columns:
            if c in row:
                values.append(str(row[c]))
            else:
                values.append("")
        self.EFH.write(self.delimiter.join(values) + "\n")

    def finalize(self) -> None:
        """
        Close file handles and create an archive if compression mode is defined.
        """
        self.NFH.close()
        self.EFH.close()
        if self.mode:
            archive_basename = f"{self.basename}.{archive_format[self.mode]}"
            archive_name = os.path.join(
                self.dirname if self.dirname else "", archive_basename
            )
            with tarfile.open(name=archive_name, mode=self.mode) as tar:
                tar.add(self.nodes_file_name, arcname=self.nodes_file_basename)
                tar.add(self.edges_file_name, arcname=self.edges_file_basename)
                if os.path.isfile(self.nodes_file_name):
                    os.remove(self.nodes_file_name)
                if os.path.isfile(self.edges_file_name):
                    os.remove(self.edges_file_name)

    @staticmethod
    def _build_export_row(data: Dict) -> Dict:
        """
        Casts all values to primitive types like str or bool according to the
        specified type in ``_column_types``. Lists become pipe delimited strings.

        Parameters
        ----------
        data: Dict
            A dictionary containing key-value pairs

        Returns
        -------
        Dict
            A dictionary containing processed key-value pairs

        """
        tidy_data = {}
        for key, value in data.items():
            new_value = remove_null(value)
            if new_value:
                tidy_data[key] = _sanitize_export(key, new_value)
        return tidy_data

    @staticmethod
    def _order_node_columns(cols: Set) -> OrderedSet:
        """
        Arrange node columns in a defined order.

        Parameters
        ----------
        cols: Set
            A set with elements in any order

        Returns
        -------
        OrderedSet
            A set with elements in a defined order

        """
        node_columns = cols.copy()
        core_columns = OrderedSet(
            ["id", "category", "name", "description", "xref", "provided_by", "synonym"]
        )
        ordered_columns = OrderedSet()
        for c in core_columns:
            if c in node_columns:
                ordered_columns.add(c)
                node_columns.remove(c)
        internal_columns = set()
        remaining_columns = node_columns.copy()
        for c in node_columns:
            if c.startswith("_"):
                internal_columns.add(c)
                remaining_columns.remove(c)
        ordered_columns.update(sorted(remaining_columns))
        ordered_columns.update(sorted(internal_columns))
        return ordered_columns

    @staticmethod
    def _order_edge_columns(cols: Set) -> OrderedSet:
        """
        Arrange edge columns in a defined order.

        Parameters
        ----------
        cols: Set
            A set with elements in any order

        Returns
        -------
        OrderedSet
            A set with elements in a defined order

        """
        edge_columns = cols.copy()
        core_columns = OrderedSet(
            [
                "id",
                "subject",
                "predicate",
                "object",
                "category",
                "relation",
                "provided_by",
            ]
        )
        ordered_columns = OrderedSet()
        for c in core_columns:
            if c in edge_columns:
                ordered_columns.add(c)
                edge_columns.remove(c)
        internal_columns = set()
        remaining_columns = edge_columns.copy()
        for c in edge_columns:
            if c.startswith("_"):
                internal_columns.add(c)
                remaining_columns.remove(c)
        ordered_columns.update(sorted(remaining_columns))
        ordered_columns.update(sorted(internal_columns))
        return ordered_columns

    def set_node_properties(self, node_properties: List) -> None:
        """
        Update node properties index with a given list.

        Parameters
        ----------
        node_properties: List
            A list of node properties

        """
        self._node_properties.update(node_properties)
        self.ordered_node_columns = TsvSink._order_node_columns(self._node_properties)

    def set_edge_properties(self, edge_properties: List) -> None:
        """
        Update edge properties index with a given list.

        Parameters
        ----------
        edge_properties: List
            A list of edge properties

        """
        self._edge_properties.update(edge_properties)
        self.ordered_edge_columns = TsvSink._order_edge_columns(self._edge_properties)



import gzip
from itertools import chain
from typing import Optional, Dict, Generator, Any
import ijson
import stringcase
from bmt import Toolkit

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.source.json_source import JsonSource
from kgx.utils.kgx_utils import get_biolink_element, format_biolink_slots

log = get_logger()


class ObographSource(JsonSource):
    """
    ObographSource is responsible for reading data as records
    from an OBO Graph JSON.
    """

    HAS_OBO_NAMESPACE = "http://www.geneontology.org/formats/oboInOwl#hasOBONamespace"
    SKOS_EXACT_MATCH = "http://www.w3.org/2004/02/skos/core#exactMatch"

    def __init__(self):
        super().__init__()
        self.toolkit = Toolkit()
        self.ecache: Dict = {}

    def parse(
        self,
        filename: str,
        format: str = "json",
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from JSON and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``json``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """
        self.set_provenance_map(kwargs)

        n = self.read_nodes(filename, compression)
        e = self.read_edges(filename, compression)
        yield from chain(n, e)

    def read_nodes(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read node records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for node records

        """
        if compression and compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for n in ijson.items(FH, "graphs.item.nodes.item"):
            yield self.read_node(n)

    def read_node(self, node: Dict) -> Dict:
        """
        Read and parse a node record.

        Parameters
        ----------
        node: Dict
            The node record

        Returns
        -------
        Dict
            The processed node

        """
        curie = self.prefix_manager.contract(node["id"])
        node_properties = {}
        if "meta" in node:
            node_properties = self.parse_meta(node["id"], node["meta"])

        fixed_node = dict()
        fixed_node["id"] = curie
        if "lbl" in node:
            fixed_node["name"] = node["lbl"]
        fixed_node["iri"] = node["id"]

        if "description" in node_properties:
            fixed_node["description"] = node_properties["description"]
        if "synonym" in node_properties:
            fixed_node["synonym"] = node_properties["synonym"]
        if "xrefs" in node_properties:
            fixed_node["xref"] = node_properties["xrefs"]
        if "subsets" in node_properties:
            fixed_node["subsets"] = node_properties["subsets"]

        if "category" not in node:
            category = self.get_category(curie, node)
            if category:
                fixed_node["category"] = [category]
            else:
                fixed_node["category"] = ["biolink:OntologyClass"]
        if "equivalent_nodes" in node_properties:
            equivalent_nodes = node_properties["equivalent_nodes"]
            fixed_node["same_as"] = equivalent_nodes
            # for n in node_properties['equivalent_nodes']:
            #     data = {'subject': fixed_node['id'], 'predicate': 'biolink:same_as',
            #     'object': n, 'relation': 'owl:sameAs'}
            #     super().load_node({'id': n, 'category': ['biolink:OntologyClass']})
            #     self.graph.add_edge(fixed_node['id'], n, **data)
        return super().read_node(fixed_node)

    def read_edges(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read edge records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for edge records

        """
        if compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for e in ijson.items(FH, "graphs.item.edges.item"):
            yield self.read_edge(e)

    def read_edge(self, edge: Dict) -> Dict:
        """
        Read and parse an edge record.

        Parameters
        ----------
        edge: Dict
            The edge record

        Returns
        -------
        Dict
            The processed edge

        """
        fixed_edge = dict()
        fixed_edge["subject"] = self.prefix_manager.contract(edge["sub"])
        if PrefixManager.is_iri(edge["pred"]):
            curie = self.prefix_manager.contract(edge["pred"])
            if curie in self.ecache:
                edge_predicate = self.ecache[curie]
            else:
                element = get_biolink_element(curie)
                if not element:
                    try:
                        mapping = self.toolkit.get_element_by_mapping(edge["pred"])
                        if mapping:
                            element = self.toolkit.get_element(mapping)
                    except ValueError as e:
                        log.error(e)

                if element:
                    edge_predicate = format_biolink_slots(element.name.replace(",", ""))
                    fixed_edge["predicate"] = edge_predicate
                else:
                    edge_predicate = "biolink:related_to"
                self.ecache[curie] = edge_predicate
            fixed_edge["predicate"] = edge_predicate
            fixed_edge["relation"] = curie
        else:
            if edge["pred"] == "is_a":
                fixed_edge["predicate"] = "biolink:subclass_of"
                fixed_edge["relation"] = "rdfs:subClassOf"
            elif edge["pred"] == "has_part":
                fixed_edge["predicate"] = "biolink:has_part"
                fixed_edge["relation"] = "BFO:0000051"
            elif edge["pred"] == "part_of":
                fixed_edge["predicate"] = "biolink:part_of"
                fixed_edge["relation"] = "BFO:0000050"
            else:
                fixed_edge["predicate"] = f"biolink:{edge['pred'].replace(' ', '_')}"
                fixed_edge["relation"] = edge["pred"]

        fixed_edge["object"] = self.prefix_manager.contract(edge["obj"])
        for x in edge.keys():
            if x not in {"sub", "pred", "obj"}:
                fixed_edge[x] = edge[x]
        return super().read_edge(fixed_edge)

    def get_category(self, curie: str, node: dict) -> Optional[str]:
        """
        Get category for a given CURIE.

        Parameters
        ----------
        curie: str
            Curie for node
        node: dict
            Node data

        Returns
        -------
        Optional[str]
            Category for the given node CURIE.

        """
        category = None
        # use meta.basicPropertyValues
        if "meta" in node and "basicPropertyValues" in node["meta"]:
            for p in node["meta"]["basicPropertyValues"]:
                if p["pred"] == self.HAS_OBO_NAMESPACE:
                    category = p["val"]
                    element = self.toolkit.get_element(category)
                    if element:
                        category = f"biolink:{stringcase.pascalcase(stringcase.snakecase(element.name))}"
                    else:
                        element = self.toolkit.get_element_by_mapping(category)
                        if element:
                            category = f"biolink:{stringcase.pascalcase(stringcase.snakecase(element.name))}"
                        else:
                            category = "biolink:OntologyClass"

        if not category or category == "biolink:OntologyClass":
            prefix = PrefixManager.get_prefix(curie)
            # TODO: the mapping should be via biolink-model lookups
            if prefix == "HP":
                category = "biolink:PhenotypicFeature"
            elif prefix == "CHEBI":
                category = "biolink:ChemicalSubstance"
            elif prefix == "MONDO":
                category = "biolink:Disease"
            elif prefix == "UBERON":
                category = "biolink:AnatomicalEntity"
            elif prefix == "SO":
                category = "biolink:SequenceFeature"
            elif prefix == "CL":
                category = "biolink:Cell"
            elif prefix == "PR":
                category = "biolink:Protein"
            elif prefix == "NCBITaxon":
                category = "biolink:OrganismalEntity"
            else:
                log.debug(
                    f"{curie} Could not find a category mapping for '{category}'; Defaulting to 'biolink:OntologyClass'"
                )
        return category

    def parse_meta(self, node: str, meta: Dict) -> Dict:
        """
        Parse 'meta' field of a node.

        Parameters
        ----------
        node: str
            Node identifier
        meta: Dict
            meta dictionary for the node

        Returns
        -------
        Dict
            A dictionary that contains 'description', 'synonyms',
            'xrefs', and 'equivalent_nodes'.

        """
        # cross species links are in meta; this needs to be parsed properly too
        # do not put assumptions in code; import as much as possible

        properties = {}
        if "definition" in meta:
            # parse 'definition' as 'description'
            description = meta["definition"]["val"]
            properties["description"] = description

        if "subsets" in meta:
            # parse 'subsets'
            subsets = meta["subsets"]
            properties["subsets"] = [
                x.split("#")[1] if "#" in x else x for x in subsets
            ]

        if "synonyms" in meta:
            # parse 'synonyms' as 'synonym'
            synonyms = [s["val"] for s in meta["synonyms"]]
            properties["synonym"] = synonyms

        if "xrefs" in meta:
            # parse 'xrefs' as 'xrefs'
            xrefs = [x["val"] for x in meta["xrefs"]]
            properties["xrefs"] = xrefs

        if "deprecated" in meta:
            # parse 'deprecated' flag
            properties["deprecated"] = meta["deprecated"]

        equivalent_nodes = []
        if "basicPropertyValues" in meta:
            # parse SKOS_EXACT_MATCH entries as 'equivalent_nodes'
            for p in meta["basicPropertyValues"]:
                if p["pred"] in {self.SKOS_EXACT_MATCH}:
                    n = self.prefix_manager.contract(p["val"])
                    if not n:
                        n = p["val"]
                    equivalent_nodes.append(n)
        properties["equivalent_nodes"] = equivalent_nodes
        return properties



import gzip
from typing import Any, Optional, Dict

import jsonstreams

from kgx.config import get_logger
from kgx.sink import Sink


log = get_logger()


class JsonSink(Sink):
    """
    JsonSink is responsible for writing data as records
    to a JSON.

    Parameters
    ----------
    filename: str
        The filename to write to
    format: str
        The file format (``json``)
    compression: Optional[str]
        The compression type (``gz``)
    kwargs: Any
        Any additional arguments

    """

    def __init__(
        self,
        filename: str,
        format: str = "json",
        compression: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.filename = filename
        if compression:
            self.compression = compression
        else:
            self.compression = None
        self.FH = jsonstreams.Stream(
            jsonstreams.Type.object, filename=filename, pretty=True, indent=4
        )
        self.NH = None
        self.EH = None

    def write_node(self, record: Dict) -> None:
        """
        Write a node record to JSON.

        Parameters
        ----------
        record: Dict
            A node record

        """
        if self.EH:
            self.EH.close()
            self.EH = None
        if not self.NH:
            self.NH = self.FH.subarray("nodes")
        self.NH.write(record)

    def write_edge(self, record: Dict) -> None:
        """
        Write an edge record to JSON.

        Parameters
        ----------
        record: Dict
            An edge record

        """
        if self.NH:
            self.NH.close()
            self.NH = None
        if not self.EH:
            self.EH = self.FH.subarray("edges")
        self.EH.write(record)

    def finalize(self) -> None:
        """
        Finalize by creating a compressed file, if needed.
        """
        if self.NH:
            self.NH.close()
        if self.EH:
            self.EH.close()
        if self.FH:
            self.FH.close()
        if self.compression:
            WH = gzip.open(f"{self.filename}.gz", "wb")
            with open(self.filename, "r") as FH:
                for line in FH.buffer:
                    WH.write(line)



from itertools import chain
from typing import Generator, Any, Dict, Optional

from kgx.config import get_graph_store_class
from kgx.graph.base_graph import BaseGraph
from kgx.source.source import Source
from kgx.utils.kgx_utils import validate_node, validate_edge, sanitize_import


class GraphSource(Source):
    """
    GraphSource is responsible for reading data as records
    from an in memory graph representation.

    The underlying store must be an instance of ``kgx.graph.base_graph.BaseGraph``
    """

    def __init__(self):
        super().__init__()
        self.graph = get_graph_store_class()()

    def parse(self, graph: BaseGraph, **kwargs: Any) -> Generator:
        """
        This method reads from a graph and yields records.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph to read from
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records read from the graph

        """
        self.graph = graph

        self.set_provenance_map(kwargs)

        nodes = self.read_nodes()
        edges = self.read_edges()
        yield from chain(nodes, edges)

    def read_nodes(self) -> Generator:
        """
        Read nodes as records from the graph.

        Returns
        -------
        Generator
            A generator for nodes

        """
        for n, data in self.graph.nodes(data=True):
            if "id" not in data:
                data["id"] = n
            node_data = validate_node(data)
            node_data = sanitize_import(node_data.copy())

            self.set_node_provenance(node_data)

            if self.check_node_filter(node_data):
                self.node_properties.update(node_data.keys())
                yield n, node_data

    def read_edges(self) -> Generator:
        """
        Read edges as records from the graph.

        Returns
        -------
        Generator
            A generator for edges

        """
        for u, v, k, data in self.graph.edges(keys=True, data=True):
            edge_data = validate_edge(data)
            edge_data = sanitize_import(edge_data.copy())

            self.set_edge_provenance(edge_data)

            if self.check_edge_filter(edge_data):
                self.node_properties.update(edge_data.keys())
                yield u, v, k, edge_data



import gzip
import ijson
from itertools import chain
from typing import Dict, Tuple, Generator, Optional, Any

from kgx.source.json_source import JsonSource


# TODO: update for TRAPI 1.0 spec


class TrapiSource(JsonSource):
    """
    TrapiSource is responsible for reading data as records
    from a TRAPI JSON.
    """

    def __init__(self):
        super().__init__()
        self._node_properties = set()
        self._edge_properties = set()

    def parse(
        self,
        filename: str,
        format: str = "json",
        compression: Optional[str] = None,
        **kwargs: Any
    ) -> Generator:
        """
        This method reads from a JSON and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``trapi-json``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """

        self.set_provenance_map(kwargs)

        n = self.read_nodes(filename, compression)
        e = self.read_edges(filename, compression)
        yield from chain(n, e)

    def read_nodes(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read node records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for node records

        """
        if compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for n in ijson.items(FH, "knowledge_graph.nodes.item"):
            yield self.load_node(n)

    def read_edges(self, filename: str, compression: Optional[str] = None) -> Generator:
        """
        Read edge records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from
        compression: Optional[str]
            The compression type

        Returns
        -------
        Generator
            A generator for edge records

        """
        if compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for e in ijson.items(FH, "knowledge_graph.edges.item"):
            yield self.load_edge(e)

    def load_node(self, node: Dict) -> Tuple[str, Dict]:
        """
        Load a node into an instance of BaseGraph

        .. Note::
            This method transformers Reasoner Std API format fields to Biolink Model fields.

        Parameters
        ----------
        node : Dict
            A node

        """

        if "type" in node and "category" not in node:
            node["category"] = node["type"]
            del node["type"]
        return super().read_node(node)

    def load_edge(self, edge: Dict) -> Tuple[str, str, str, Dict]:
        """
        Load an edge into an instance of BaseGraph

        .. Note::
            This methods transformers Reasoner Std API format fields to Biolink Model fields.

        Parameters
        ----------
        edge : Dict
            An edge

        """
        if "source_id" in edge:
            edge["subject"] = edge["source_id"]
        if "target_id" in edge:
            edge["object"] = edge["target_id"]
        if "relation_label" in edge:
            edge["predicate"] = edge["relation_label"][0]
        return super().read_edge(edge)



import importlib
import re
import time
import uuid
from enum import Enum
from typing import List, Dict, Set, Optional, Any, Union
import stringcase
from linkml_runtime.linkml_model.meta import (
    TypeDefinitionName,
    ElementName,
    SlotDefinition,
    ClassDefinition,
    TypeDefinition,
    Element,
)
from bmt import Toolkit
from cachetools import LRUCache
import pandas as pd
import numpy as np
from prefixcommons.curie_util import contract_uri
from prefixcommons.curie_util import expand_uri

from kgx.config import get_logger, get_jsonld_context, get_biolink_model_schema
from kgx.graph.base_graph import BaseGraph

curie_lookup_service = None
cache = None

log = get_logger()

DEFAULT_NODE_CATEGORY = "biolink:NamedThing"
DEFAULT_EDGE_PREDICATE = "biolink:related_to"
CORE_NODE_PROPERTIES = {"id", "name"}
CORE_EDGE_PROPERTIES = {"id", "subject", "predicate", "object", "type"}

LIST_DELIMITER = "|"


class GraphEntityType(Enum):
    GRAPH = "graph"
    NODE = "node"
    EDGE = "edge"


# Biolink 2.0 "Knowledge Source" association slots,
# including the deprecated 'provided_by' slot

provenance_slot_types = {
    "knowledge_source": list,
    "primary_knowledge_source": str,
    "original_knowledge_source": str,
    "aggregator_knowledge_source": list,
    "supporting_data_source": list,
    "provided_by": list,
}

column_types = {
    "publications": list,
    "qualifiers": list,
    "category": list,
    "synonym": list,
    "same_as": list,
    "negated": bool,
    "xrefs": list,
}

column_types.update(provenance_slot_types)

knowledge_provenance_properties = set(provenance_slot_types.keys())

extension_types = {"csv": ",", "tsv": "\t", "csv:neo4j": ",", "tsv:neo4j": "\t"}

archive_read_mode = {"tar": "r", "tar.gz": "r:gz", "tar.bz2": "r:bz2"}
archive_write_mode = {"tar": "w", "tar.gz": "w:gz", "tar.bz2": "w:bz2"}

archive_format = {
    "r": "tar",
    "r:gz": "tar.gz",
    "r:bz2": "tar.bz2",
    "w": "tar",
    "w:gz": "tar.gz",
    "w:bz2": "tar.bz2",
}

is_provenance_property_multivalued = {
    "knowledge_source": True,
    "primary_knowledge_source": False,
    "original_knowledge_source": False,
    "aggregator_knowledge_source": True,
    "supporting_data_source": True,
    "provided_by": True,
}

is_property_multivalued = {
    "id": False,
    "subject": False,
    "object": False,
    "predicate": False,
    "description": False,
    "synonym": True,
    "in_taxon": False,
    "same_as": True,
    "name": False,
    "has_evidence": False,
    "category": True,
    "publications": True,
    "type": False,
    "relation": False,
}

is_property_multivalued.update(is_provenance_property_multivalued)


def camelcase_to_sentencecase(s: str) -> str:
    """
    Convert CamelCase to sentence case.

    Parameters
    ----------
    s: str
        Input string in CamelCase

    Returns
    -------
    str
        string in sentence case form

    """
    return stringcase.sentencecase(s).lower()


def snakecase_to_sentencecase(s: str) -> str:
    """
    Convert snake_case to sentence case.

    Parameters
    ----------
    s: str
        Input string in snake_case

    Returns
    -------
    str
        string in sentence case form

    """
    return stringcase.sentencecase(s).lower()


def sentencecase_to_snakecase(s: str) -> str:
    """
    Convert sentence case to snake_case.

    Parameters
    ----------
    s: str
        Input string in sentence case

    Returns
    -------
    str
        string in snake_case form

    """
    return stringcase.snakecase(s).lower()


def sentencecase_to_camelcase(s: str) -> str:
    """
    Convert sentence case to CamelCase.

    Parameters
    ----------
    s: str
        Input string in sentence case

    Returns
    -------
    str
        string in CamelCase form

    """
    return stringcase.pascalcase(stringcase.snakecase(s))


def format_biolink_category(s: str) -> str:
    """
    Convert a sentence case Biolink category name to
    a proper Biolink CURIE with the category itself
    in CamelCase form.

    Parameters
    ----------
    s: str
        Input string in sentence case

    Returns
    -------
    str
        a proper Biolink CURIE
    """
    if re.match("biolink:.+", s):
        return s
    else:
        formatted = sentencecase_to_camelcase(s)
        return f"biolink:{formatted}"


def format_biolink_slots(s: str) -> str:
    if re.match("biolink:.+", s):
        return s
    else:
        formatted = sentencecase_to_snakecase(s)
        return f"biolink:{formatted}"


def contract(
    uri: str, prefix_maps: Optional[List[Dict]] = None, fallback: bool = True
) -> str:
    """
    Contract a given URI to a CURIE, based on mappings from `prefix_maps`.
    If no prefix map is provided then will use defaults from prefixcommons-py.

    This method will return the URI as the CURIE if there is no mapping found.

    Parameters
    ----------
    uri: str
        A URI
    prefix_maps: Optional[List[Dict]]
        A list of prefix maps to use for mapping
    fallback: bool
        Determines whether to fallback to default prefix mappings, as determined
        by `prefixcommons.curie_util`, when URI prefix is not found in `prefix_maps`.

    Returns
    -------
    str
        A CURIE corresponding to the URI

    """
    curie = uri
    default_curie_maps = [
        get_jsonld_context("monarch_context"),
        get_jsonld_context("obo_context"),
    ]
    if prefix_maps:
        curie_list = contract_uri(uri, prefix_maps)
        if len(curie_list) == 0:
            if fallback:
                curie_list = contract_uri(uri, default_curie_maps)
                if curie_list:
                    curie = curie_list[0]
        else:
            curie = curie_list[0]
    else:
        curie_list = contract_uri(uri, default_curie_maps)
        if len(curie_list) > 0:
            curie = curie_list[0]

    return curie


def expand(
    curie: str, prefix_maps: Optional[List[dict]] = None, fallback: bool = True
) -> str:
    """
    Expand a given CURIE to an URI, based on mappings from `prefix_map`.

    This method will return the CURIE as the IRI if there is no mapping found.

    Parameters
    ----------
    curie: str
        A CURIE
    prefix_maps: Optional[List[dict]]
        A list of prefix maps to use for mapping
    fallback: bool
        Determines whether to fallback to default prefix mappings, as determined
        by `prefixcommons.curie_util`, when CURIE prefix is not found in `prefix_maps`.

    Returns
    -------
    str
        A URI corresponding to the CURIE

    """
    default_curie_maps = [
        get_jsonld_context("monarch_context"),
        get_jsonld_context("obo_context"),
    ]
    if prefix_maps:
        uri = expand_uri(curie, prefix_maps)
        if uri == curie and fallback:
            uri = expand_uri(curie, default_curie_maps)
    else:
        uri = expand_uri(curie, default_curie_maps)

    return uri


_default_toolkit = None

# TODO: not sure how threadsafe this simple-minded Toolkit cache is
_toolkit_versions: Dict[str, Toolkit] = dict()


def get_toolkit(biolink_release: Optional[str] = None) -> Toolkit:
    """
    Get an instance of bmt.Toolkit
    If there no instance defined, then one is instantiated and returned.

    Parameters
    ----------
    biolink_release: Optional[str]
        URL to (Biolink) Model Schema to be used for validated (default: None, use default Biolink Model Toolkit schema)

    """
    global _default_toolkit, _toolkit_versions
    if biolink_release:
        if biolink_release in _toolkit_versions:
            toolkit = _toolkit_versions[biolink_release]
        else:
            schema = get_biolink_model_schema(biolink_release)
            toolkit = Toolkit(schema=schema)
            _toolkit_versions[biolink_release] = toolkit
    else:
        if _default_toolkit is None:
            _default_toolkit = Toolkit()
        toolkit = _default_toolkit
        biolink_release = toolkit.get_model_version()
        if biolink_release not in _toolkit_versions:
            _toolkit_versions[biolink_release] = toolkit

    return toolkit


def generate_edge_key(s: str, edge_predicate: str, o: str) -> str:
    """
    Generates an edge key based on a given subject, predicate, and object.

    Parameters
    ----------
    s: str
        Subject
    edge_predicate: str
        Edge label
    o: str
        Object

    Returns
    -------
    str
        Edge key as a string

    """
    return "{}-{}-{}".format(s, edge_predicate, o)


def get_curie_lookup_service():
    """
    Get an instance of kgx.curie_lookup_service.CurieLookupService

    Returns
    -------
    kgx.curie_lookup_service.CurieLookupService
        An instance of ``CurieLookupService``

    """
    global curie_lookup_service
    if curie_lookup_service is None:
        from kgx.curie_lookup_service import CurieLookupService

        curie_lookup_service = CurieLookupService()
    return curie_lookup_service


def get_cache(maxsize=10000):
    """
    Get an instance of cachetools.cache

    Parameters
    ----------
    maxsize: int
        The max size for the cache (``10000``, by default)

    Returns
    -------
    cachetools.cache
        An instance of cachetools.cache

    """
    global cache
    if cache is None:
        cache = LRUCache(maxsize)
    return cache


def current_time_in_millis():
    """
    Get current time in milliseconds.

    Returns
    -------
    int
        Time in milliseconds

    """
    return int(round(time.time() * 1000))


def get_prefix_prioritization_map() -> Dict[str, List]:
    """
    Get prefix prioritization map as defined in Biolink Model.

    Returns
    -------
    Dict[str, List]

    """
    toolkit = get_toolkit()
    prefix_prioritization_map = {}
    # TODO: Lookup via Biolink CURIE should be supported in bmt
    descendants = toolkit.get_descendants("named thing")
    descendants.append("named thing")
    for d in descendants:
        element = toolkit.get_element(d)
        if element and "id_prefixes" in element:
            prefixes = element.id_prefixes
            key = format_biolink_category(element.name)
            prefix_prioritization_map[key] = prefixes
    return prefix_prioritization_map


def get_biolink_element(name) -> Optional[Element]:
    """
    Get Biolink element for a given name, where name can be a class, slot, or relation.

    Parameters
    ----------
    name: str
        The name

    Returns
    -------
    Optional[linkml_model.meta.Element]
        An instance of linkml_model.meta.Element

    """
    toolkit = get_toolkit()
    element = toolkit.get_element(name)
    return element


def get_biolink_ancestors(name: str):
    """
    Get ancestors for a given Biolink class.

    Parameters
    ----------
    name: str

    Returns
    -------
    List
        A list of ancestors

    """
    toolkit = get_toolkit()
    ancestors_mixins = toolkit.get_ancestors(name, formatted=True, mixin=True)
    return ancestors_mixins


def get_biolink_property_types() -> Dict:
    """
    Get all Biolink property types.
    This includes both node and edges properties.

    Returns
    -------
    Dict
        A dict containing all Biolink property and their types

    """
    toolkit = get_toolkit()
    types = {}
    node_properties = toolkit.get_all_node_properties(formatted=True)
    edge_properties = toolkit.get_all_edge_properties(formatted=True)

    for p in node_properties:
        property_type = get_type_for_property(p)
        types[p] = property_type

    for p in edge_properties:
        property_type = get_type_for_property(p)
        types[p] = property_type

    # TODO: this should be moved to biolink model
    types["biolink:predicate"] = "uriorcurie"
    types["biolink:edge_label"] = "uriorcurie"
    return types


def get_type_for_property(p: str) -> str:
    """
    Get type for a property.

    TODO: Move this to biolink-model-default_toolkit

    Parameters
    ----------
    p: str

    Returns
    -------
    str
        The type for a given property

    """
    toolkit = get_toolkit()
    e = toolkit.get_element(p)
    t = "xsd:string"
    if e:
        if isinstance(e, ClassDefinition):
            t = "uriorcurie"
        elif isinstance(e, TypeDefinition):
            t = e.uri
        else:
            r = e.range
            if isinstance(r, SlotDefinition):
                t = r.range
                t = get_type_for_property(t)
            elif isinstance(r, TypeDefinitionName):
                t = get_type_for_property(r)
            elif isinstance(r, ElementName):
                t = get_type_for_property(r)
            else:
                t = "xsd:string"
    return t


def prepare_data_dict(d1: Dict, d2: Dict, preserve: bool = True) -> Dict:
    """
    Given two dict objects, make a new dict object that is the intersection of the two.

    If a key is known to be multivalued then it's value is converted to a list.
    If a key is already multivalued then it is updated with new values.
    If a key is single valued, and a new unique value is found then the existing value is
    converted to a list and the new value is appended to this list.

    Parameters
    ----------
    d1: Dict
        Dict object
    d2: Dict
        Dict object
    preserve: bool
        Whether or not to preserve values for conflicting keys

    Returns
    -------
    Dict
        The intersection of d1 and d2

    """
    new_data = {}
    for key, value in d2.items():
        if isinstance(value, (list, set, tuple)):
            new_value = [x for x in value]
        else:
            new_value = value

        if key in is_property_multivalued:
            if is_property_multivalued[key]:
                # value for key is supposed to be multivalued
                if key in d1:
                    # key is in data
                    if isinstance(d1[key], (list, set, tuple)):
                        # existing key has value type list
                        new_data[key] = d1[key]
                        if isinstance(new_value, (list, set, tuple)):
                            new_data[key] += [
                                x for x in new_value if x not in new_data[key]
                            ]
                        else:
                            if new_value not in new_data[key]:
                                new_data[key].append(new_value)
                    else:
                        if key in CORE_NODE_PROPERTIES or key in CORE_EDGE_PROPERTIES:
                            log.debug(
                                f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                            )
                        else:
                            # existing key does not have value type list; converting to list
                            new_data[key] = [d1[key]]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [
                                    x for x in new_value if x not in new_data[key]
                                ]
                            else:
                                if new_value not in new_data[key]:
                                    new_data[key].append(new_value)
                else:
                    # key is not in data; adding
                    if isinstance(new_value, (list, set, tuple)):
                        new_data[key] = [x for x in new_value]
                    else:
                        new_data[key] = [new_value]
            else:
                # key is not multivalued; adding/replacing as-is
                if key in d1:
                    if isinstance(d1[key], (list, set, tuple)):
                        new_data[key] = d1[key]
                        if isinstance(new_value, (list, set, tuple)):
                            new_data[key] += [x for x in new_value]
                        else:
                            new_data[key].append(new_value)
                    else:
                        if key in CORE_NODE_PROPERTIES or key in CORE_EDGE_PROPERTIES:
                            log.debug(
                                f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                            )
                        else:
                            if preserve:
                                new_data[key] = [d1[key]]
                                if isinstance(new_value, (list, set, tuple)):
                                    new_data[key] += [
                                        x for x in new_value if x not in new_data[key]
                                    ]
                                else:
                                    new_data[key].append(new_value)
                            else:
                                new_data[key] = new_value
                else:
                    new_data[key] = new_value
        else:
            # treating key as multivalued
            if key in d1:
                # key is in data
                if key in CORE_NODE_PROPERTIES or key in CORE_EDGE_PROPERTIES:
                    log.debug(
                        f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                    )
                else:
                    if isinstance(d1[key], (list, set, tuple)):
                        # existing key has value type list
                        new_data[key] = d1[key]
                        if isinstance(new_value, (list, set, tuple)):
                            new_data[key] += [
                                x for x in new_value if x not in new_data[key]
                            ]
                        else:
                            new_data[key].append(new_value)
                    else:
                        # existing key does not have value type list; converting to list
                        if preserve:
                            new_data[key] = [d1[key]]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [
                                    x for x in new_value if x not in new_data[key]
                                ]
                            else:
                                new_data[key].append(new_value)
                        else:
                            new_data[key] = new_value
            else:
                new_data[key] = new_value

    for key, value in d1.items():
        if key not in new_data:
            new_data[key] = value
    return new_data


def apply_filters(
    graph: BaseGraph,
    node_filters: Dict[str, Union[str, Set]],
    edge_filters: Dict[str, Union[str, Set]],
) -> None:
    """
    Apply filters to graph and remove nodes and edges that
    do not pass given filters.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    node_filters: Dict[str, Union[str, Set]]
        Node filters
    edge_filters: Dict[str, Union[str, Set]]
        Edge filters

    """
    apply_node_filters(graph, node_filters)
    apply_edge_filters(graph, edge_filters)


def apply_node_filters(
    graph: BaseGraph, node_filters: Dict[str, Union[str, Set]]
) -> None:
    """
    Apply filters to graph and remove nodes that do not pass given filters.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    node_filters: Dict[str, Union[str, Set]]
        Node filters

    """
    nodes_to_remove = []
    for node, node_data in graph.nodes(data=True):
        pass_filter = True
        for k, v in node_filters.items():
            if k == "category":
                if not any(x in node_data[k] for x in v):
                    pass_filter = False
        if not pass_filter:
            nodes_to_remove.append(node)

    for node in nodes_to_remove:
        # removing node that fails category filter
        log.debug(f"Removing node {node}")
        graph.remove_node(node)


def apply_edge_filters(
    graph: BaseGraph, edge_filters: Dict[str, Union[str, Set]]
) -> None:
    """
    Apply filters to graph and remove edges that do not pass given filters.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    edge_filters: Dict[str, Union[str, Set]]
        Edge filters

    """
    edges_to_remove = []
    for subject_node, object_node, key, data in graph.edges(keys=True, data=True):
        pass_filter = True
        for k, v in edge_filters.items():
            if k == "predicate":
                if data[k] not in v:
                    pass_filter = False
            elif k == "relation":
                if data[k] not in v:
                    pass_filter = False
        if not pass_filter:
            edges_to_remove.append((subject_node, object_node, key))

    for edge in edges_to_remove:
        # removing edge that fails edge filters
        log.debug(f"Removing edge {edge}")
        graph.remove_edge(edge[0], edge[1], edge[2])


def validate_node(node: Dict) -> Dict:
    """
    Given a node as a dictionary, check for required properties.
    This method will return the node dictionary with default
    assumptions applied, if any.

    Parameters
    ----------
    node: Dict
        A node represented as a dict

    Returns
    -------
    Dict
        A node represented as a dict, with default assumptions applied.

    """
    if len(node) == 0:
        log.debug(f"Empty node encountered: {node}")
    else:
        if "id" not in node:
            raise KeyError(f"node does not have 'id' property: {node}")
        if "name" not in node:
            log.debug(f"node does not have 'name' property: {node}")
        if "category" not in node:
            log.debug(
                f"node does not have 'category' property: {node}\nUsing {DEFAULT_NODE_CATEGORY} as default"
            )
            node["category"] = [DEFAULT_NODE_CATEGORY]
    return node


def validate_edge(edge: Dict) -> Dict:
    """
    Given an edge as a dictionary, check for required properties.
    This method will return the edge dictionary with default
    assumptions applied, if any.

    Parameters
    ----------
    edge: Dict
        An edge represented as a dict

    Returns
    -------
    Dict
        An edge represented as a dict, with default assumptions applied.
    """
    if "subject" not in edge:
        raise KeyError(f"edge does not have 'subject' property: {edge}")
    if "predicate" not in edge:
        raise KeyError(f"edge does not have 'predicate' property: {edge}")
    if "object" not in edge:
        raise KeyError(f"edge does not have 'object' property: {edge}")
    return edge


def generate_uuid():
    """
    Generates a UUID.

    Returns
    -------
    str
        A UUID

    """
    return f"urn:uuid:{uuid.uuid4()}"


def generate_edge_identifiers(graph: BaseGraph):
    """
    Generate unique identifiers for edges in a graph that do not
    have an ``id`` field.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph

    """
    for u, v, data in graph.edges(data=True):
        if "id" not in data:
            data["id"] = generate_uuid()


def sanitize_import(data: Dict) -> Dict:
    """
    Sanitize key-value pairs in dictionary.

    Parameters
    ----------
    data: Dict
        A dictionary containing key-value pairs

    Returns
    -------
    Dict
        A dictionary containing processed key-value pairs

    """
    tidy_data = {}
    for key, value in data.items():
        new_value = remove_null(value)
        if new_value is not None:
            tidy_data[key] = _sanitize_import(key, new_value)
    return tidy_data


def _sanitize_import(key: str, value: Any) -> Any:
    """
    Sanitize value for a key for the purpose of import.

    Parameters
    ----------
    key: str
        Key corresponding to a node/edge property
    value: Any
        Value corresponding to the key

    Returns
    -------
    value: Any
        Sanitized value

    """
    new_value: Any
    if key in column_types:
        if column_types[key] == list:
            if isinstance(value, (list, set, tuple)):
                value = [
                    v.replace("\n", " ").replace("\t", " ") if isinstance(v, str) else v
                    for v in value
                ]
                new_value = list(value)
            elif isinstance(value, str):
                value = value.replace("\n", " ").replace("\t", " ")
                new_value = [x for x in value.split(LIST_DELIMITER) if x]
            else:
                new_value = [str(value).replace("\n", " ").replace("\t", " ")]
        elif column_types[key] == bool:
            try:
                new_value = bool(value)
            except:
                new_value = False
        elif isinstance(value, (str, float)):
            new_value = value
        else:
            new_value = str(value).replace("\n", " ").replace("\t", " ")
    else:
        if isinstance(value, (list, set, tuple)):
            value = [
                v.replace("\n", " ").replace("\t", " ") if isinstance(v, str) else v
                for v in value
            ]
            new_value = list(value)
        elif isinstance(value, str):
            if LIST_DELIMITER in value:
                value = value.replace("\n", " ").replace("\t", " ")
                new_value = [x for x in value.split(LIST_DELIMITER) if x]
            else:
                new_value = value.replace("\n", " ").replace("\t", " ")
        elif isinstance(value, bool):
            try:
                new_value = bool(value)
            except:
                new_value = False
        elif isinstance(value, (str, float)):
            new_value = value
        else:
            new_value = str(value).replace("\n", " ").replace("\t", " ")
    return new_value


def _sanitize_export(key: str, value: Any) -> Any:
    """
    Sanitize value for a key for the purpose of export.

    Parameters
    ----------
    key: str
        Key corresponding to a node/edge property
    value: Any
        Value corresponding to the key

    Returns
    -------
    value: Any
        Sanitized value

    """
    new_value: Any
    if key in column_types:
        if column_types[key] == list:
            if isinstance(value, (list, set, tuple)):
                value = [
                    v.replace("\n", " ").replace('\\"', "").replace("\t", " ")
                    if isinstance(v, str)
                    else v
                    for v in value
                ]
                new_value = LIST_DELIMITER.join([str(x) for x in value])
            else:
                new_value = (
                    str(value).replace("\n", " ").replace('\\"', "").replace("\t", " ")
                )
        elif column_types[key] == bool:
            try:
                new_value = bool(value)
            except:
                new_value = False
        else:
            new_value = (
                str(value).replace("\n", " ").replace('\\"', "").replace("\t", " ")
            )
    else:
        if type(value) == list:
            new_value = LIST_DELIMITER.join([str(x) for x in value])
            new_value = (
                new_value.replace("\n", " ").replace('\\"', "").replace("\t", " ")
            )
            column_types[key] = list
        elif type(value) == bool:
            try:
                new_value = bool(value)
                column_types[key] = bool
            except:
                new_value = False
        else:
            new_value = (
                str(value).replace("\n", " ").replace('\\"', "").replace("\t", " ")
            )
    return new_value


def _build_export_row(data: Dict) -> Dict:
    """
    Casts all values to primitive types like str or bool according to the
    specified type in ``_column_types``. Lists become pipe delimited strings.

    Parameters
    ----------
    data: Dict
        A dictionary containing key-value pairs

    Returns
    -------
    Dict
        A dictionary containing processed key-value pairs

    """
    tidy_data = {}
    for key, value in data.items():
        new_value = remove_null(value)
        if new_value:
            tidy_data[key] = _sanitize_export(key, new_value)
    return tidy_data


def remove_null(input: Any) -> Any:
    """
    Remove any null values from input.

    Parameters
    ----------
    input: Any
        Can be a str, list or dict

    Returns
    -------
    Any
        The input without any null values

    """
    new_value: Any = None
    if isinstance(input, (list, set, tuple)):
        # value is a list, set or a tuple
        new_value = []
        for v in input:
            x = remove_null(v)
            if x:
                new_value.append(x)
    elif isinstance(input, dict):
        # value is a dict
        new_value = {}
        for k, v in input.items():
            x = remove_null(v)
            if x:
                new_value[k] = x
    elif isinstance(input, str):
        # value is a str
        if not is_null(input):
            new_value = input
    else:
        if not is_null(input):
            new_value = input
    return new_value


def is_null(item: Any) -> bool:
    """
    Checks if a given item is null or correspond to null.

    This method checks for: ``None``, ``numpy.nan``, ``pandas.NA``,
    ``pandas.NaT``, and ` `

    Parameters
    ----------
    item: Any
        The item to check

    Returns
    -------
    bool
        Whether the given item is null or not

    """
    null_values = {np.nan, pd.NA, pd.NaT, None, "", " "}
    return item in null_values


def apply_graph_operations(graph: BaseGraph, operations: List) -> None:
    """
    Apply graph operations to a given graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        An instance of BaseGraph
    operations: List
        A list of graph operations with configuration

    """
    for operation in operations:
        op_name = operation["name"]
        op_args = operation["args"]
        module_name = ".".join(op_name.split(".")[0:-1])
        function_name = op_name.split(".")[-1]
        f = getattr(importlib.import_module(module_name), function_name)
        f(graph, **op_args)



import gzip
from collections import OrderedDict
from typing import Optional, Union, Tuple, Any, Dict

import rdflib
from linkml_runtime.linkml_model.meta import Element, ClassDefinition, SlotDefinition
from rdflib import URIRef, Literal, Namespace, RDF
from rdflib.plugins.serializers.nt import _nt_row
from rdflib.term import _is_valid_uri

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.sink.sink import Sink
from kgx.utils.kgx_utils import (
    get_toolkit,
    sentencecase_to_camelcase,
    get_biolink_ancestors,
    sentencecase_to_snakecase,
    generate_uuid,
    get_biolink_property_types,
)
from kgx.utils.rdf_utils import process_predicate

log = get_logger()

property_mapping: OrderedDict = OrderedDict()
reverse_property_mapping: OrderedDict = OrderedDict()


class RdfSink(Sink):
    """
    RdfSink is responsible for writing data as records
    to an RDF serialization.

    .. note::
        Currently only RDF N-Triples serialization is supported.

    Parameters
    ----------
    filename: str
        The filename to write to
    format: str
        The file format (``nt``)
    compression: str
        The compression type (``gz``)
    reify_all_edges: bool
        Whether or not to reify all the edges
    kwargs: Any
        Any additional arguments

    """

    def __init__(
        self,
        filename: str,
        format: str = "nt",
        compression: Optional[bool] = None,
        reify_all_edges: bool = False,
        **kwargs: Any,
    ):
        super().__init__()
        if format not in {"nt"}:
            raise ValueError(f"Only RDF N-Triples ('nt') serialization supported.")
        self.DEFAULT = Namespace(self.prefix_manager.prefix_map[""])
        # self.OBO = Namespace('http://purl.obolibrary.org/obo/')
        self.OBAN = Namespace(self.prefix_manager.prefix_map["OBAN"])
        self.PMID = Namespace(self.prefix_manager.prefix_map["PMID"])
        self.BIOLINK = Namespace(self.prefix_manager.prefix_map["biolink"])
        self.toolkit = get_toolkit()
        self.reverse_predicate_mapping = {}
        self.property_types = get_biolink_property_types()
        self.cache = {}
        self.reify_all_edges = reify_all_edges
        self.reification_types = {
            RDF.Statement,
            self.BIOLINK.Association,
            self.OBAN.association,
        }
        if compression == "gz":
            f = gzip.open(filename, "wb")
        else:
            f = open(filename, "wb")
        self.FH = f
        self.encoding = "ascii"

    def set_reverse_predicate_mapping(self, m: Dict) -> None:
        """
        Set reverse predicate mappings.

        Use this method to update mappings for predicates that are
        not in Biolink Model.

        Parameters
        ----------
        m: Dict
            A dictionary where the keys are property names and values
            are their corresponding IRI.

        """
        for k, v in m.items():
            self.reverse_predicate_mapping[v] = URIRef(k)

    def set_property_types(self, m: Dict) -> None:
        """
        Set export type for properties that are not in
        Biolink Model.

        Parameters
        ----------
        m: Dict
            A dictionary where the keys are property names and values
            are their corresponding types.

        """
        for k, v in m.items():
            (element_uri, canonical_uri, predicate, property_name) = process_predicate(
                self.prefix_manager, k
            )
            if element_uri:
                key = element_uri
            elif predicate:
                key = predicate
            else:
                key = property_name
            self.property_types[key] = v

    def write_node(self, record: Dict) -> None:
        """
        Write a node record as triples.

        Parameters
        ----------
        record: Dict
            A node record

        """
        for k, v in record.items():
            if k in {"id", "iri"}:
                continue
            (
                element_uri,
                canonical_uri,
                predicate,
                property_name,
            ) = self.process_predicate(k)
            if element_uri is None:
                # not a biolink predicate
                if k in self.reverse_predicate_mapping:
                    prop_uri = self.reverse_predicate_mapping[k]
                    # prop_uri = self.prefix_manager.contract(prop_uri)
                else:
                    prop_uri = k
            else:
                prop_uri = canonical_uri if canonical_uri else element_uri
            prop_type = self._get_property_type(prop_uri)
            log.debug(f"prop {k} has prop_uri {prop_uri} and prop_type {prop_type}")
            prop_uri = self.uriref(prop_uri)
            if isinstance(v, (list, set, tuple)):
                for x in v:
                    value_uri = self._prepare_object(k, prop_type, x)
                    self._write_triple(self.uriref(record["id"]), prop_uri, value_uri)
            else:
                value_uri = self._prepare_object(k, prop_type, v)
                self._write_triple(self.uriref(record["id"]), prop_uri, value_uri)

    def _write_triple(self, s: URIRef, p: URIRef, o: Union[URIRef, Literal]) -> None:
        """
        Serialize a triple.

        Parameters
        ----------
        s: rdflib.URIRef
            The subject
        p: rdflib.URIRef
            The predicate
        o: Union[rdflib.URIRef, rdflib.Literal]
            The object

        """
        self.FH.write(_nt_row((s, p, o)).encode(self.encoding, "_rdflib_nt_escape"))

    def write_edge(self, record: Dict) -> None:
        """
        Write an edge record as triples.

        Parameters
        ----------
        record: Dict
            An edge record

        """
        ecache = []
        associations = set(
            [self.prefix_manager.contract(x) for x in self.reification_types]
        )
        associations.update(
            [str(x) for x in set(self.toolkit.get_all_associations(formatted=True))]
        )
        if self.reify_all_edges:
            reified_node = self.reify(record["subject"], record["object"], record)
            s = reified_node["subject"]
            p = reified_node["predicate"]
            o = reified_node["object"]
            ecache.append((s, p, o))
            n = reified_node["id"]
            for prop, value in reified_node.items():
                if prop in {"id", "association_id", "edge_key"}:
                    continue
                (
                    element_uri,
                    canonical_uri,
                    predicate,
                    property_name,
                ) = self.process_predicate(prop)
                if element_uri:
                    prop_uri = canonical_uri if canonical_uri else element_uri
                else:
                    if prop in self.reverse_predicate_mapping:
                        prop_uri = self.reverse_predicate_mapping[prop]
                        # prop_uri = self.prefix_manager.contract(prop_uri)
                    else:
                        prop_uri = predicate
                prop_type = self._get_property_type(prop)
                log.debug(
                    f"prop {prop} has prop_uri {prop_uri} and prop_type {prop_type}"
                )
                prop_uri = self.uriref(prop_uri)
                if isinstance(value, list):
                    for x in value:
                        value_uri = self._prepare_object(prop, prop_type, x)
                        self._write_triple(URIRef(n), prop_uri, value_uri)
                else:
                    value_uri = self._prepare_object(prop, prop_type, value)
                    self._write_triple(URIRef(n), prop_uri, value_uri)
        else:
            if (
                ("type" in record and record["type"] in associations)
                or (
                    "association_type" in record
                    and record["association_type"] in associations
                )
                or ("category" in record and any(record["category"]) in associations)
            ):
                reified_node = self.reify(record["subject"], record["object"], record)
                s = reified_node["subject"]
                p = reified_node["predicate"]
                o = reified_node["object"]
                ecache.append((s, p, o))
                n = reified_node["id"]
                for prop, value in reified_node.items():
                    if prop in {"id", "association_id", "edge_key"}:
                        continue
                    (
                        element_uri,
                        canonical_uri,
                        predicate,
                        property_name,
                    ) = self.process_predicate(prop)
                    if element_uri:
                        prop_uri = canonical_uri if canonical_uri else element_uri
                    else:
                        if prop in self.reverse_predicate_mapping:
                            prop_uri = self.reverse_predicate_mapping[prop]
                            # prop_uri = self.prefix_manager.contract(prop_uri)
                        else:
                            prop_uri = predicate
                    prop_type = self._get_property_type(prop)
                    prop_uri = self.uriref(prop_uri)
                    if isinstance(value, list):
                        for x in value:
                            value_uri = self._prepare_object(prop, prop_type, x)
                            self._write_triple(URIRef(n), prop_uri, value_uri)
                    else:
                        value_uri = self._prepare_object(prop, prop_type, value)
                        self._write_triple(URIRef(n), prop_uri, value_uri)
            else:
                s = self.uriref(record["subject"])
                p = self.uriref(record["predicate"])
                o = self.uriref(record["object"])
                self._write_triple(s, p, o)
        for t in ecache:
            self._write_triple(t[0], t[1], t[2])

    def uriref(self, identifier: str) -> URIRef:
        """
        Generate a rdflib.URIRef for a given string.

        Parameters
        ----------
        identifier: str
            Identifier as string.

        Returns
        -------
        rdflib.URIRef
            URIRef form of the input ``identifier``

        """
        if identifier.startswith("urn:uuid:"):
            uri = identifier
        elif identifier in reverse_property_mapping:
            # identifier is a property
            uri = reverse_property_mapping[identifier]
        else:
            # identifier is an entity
            fixed_identifier = identifier
            if fixed_identifier.startswith(":"):
                # TODO: this should be handled upstream by prefixcommons-py
                fixed_identifier = fixed_identifier.replace(":", "", 1)
            if " " in identifier:
                fixed_identifier = fixed_identifier.replace(" ", "_")

            if self.prefix_manager.is_curie(fixed_identifier):
                uri = self.prefix_manager.expand(fixed_identifier)
                if fixed_identifier == uri:
                    uri = self.DEFAULT.term(fixed_identifier)
            elif self.prefix_manager.is_iri(fixed_identifier):
                uri = fixed_identifier
            else:
                uri = self.DEFAULT.term(fixed_identifier)
            # if identifier == uri:
            #     if PrefixManager.is_curie(identifier):
            #         identifier = identifier.replace(':', '_')
        return URIRef(uri)

    def _prepare_object(
        self, prop: str, prop_type: str, value: Any
    ) -> rdflib.term.Identifier:
        """
        Prepare the object of a triple.

        Parameters
        ----------
        prop: str
            property name
        prop_type: str
            property type
        value: Any
            property value

        Returns
        -------
        rdflib.term.Identifier
            An instance of rdflib.term.Identifier

        """
        if prop_type == "uriorcurie" or prop_type == "xsd:anyURI":
            if isinstance(value, str) and PrefixManager.is_curie(value):
                o = self.uriref(value)
            elif isinstance(value, str) and PrefixManager.is_iri(value):
                if _is_valid_uri(value):
                    o = URIRef(value)
                else:
                    o = Literal(value)
            else:
                o = Literal(value)
        elif prop_type.startswith("xsd"):
            o = Literal(value, datatype=self.prefix_manager.expand(prop_type))
        else:
            o = Literal(value, datatype=self.prefix_manager.expand("xsd:string"))
        return o

    def _get_property_type(self, p: str) -> str:
        """
        Get type for a given property name.

        Parameters
        ----------
        p: str
            property name

        Returns
        -------
        str
            The type for property name

        """
        # TODO: this should be properly defined in the model
        default_uri_types = {
            "biolink:type",
            "biolink:category",
            "biolink:subject",
            "biolink:object",
            "biolink:relation",
            "biolink:predicate",
            "rdf:type",
            "rdf:subject",
            "rdf:predicate",
            "rdf:object",
        }
        if p in default_uri_types:
            t = "uriorcurie"
        else:
            if p in self.property_types:
                t = self.property_types[p]
            elif f":{p}" in self.property_types:
                t = self.property_types[f":{p}"]
            elif f"biolink:{p}" in self.property_types:
                t = self.property_types[f"biolink:{p}"]
            else:
                t = "xsd:string"
            # if value:
            #     if isinstance(value, (list, set, tuple)):
            #         x = value[0]
            #         if self.graph.has_node(x):
            #             t = 'uriorcurie'
            #         else:
            #             t = 'xsd:string'
            #     else:
            #         if self.graph.has_node(value):
            #             t = 'uriorcurie'
            #         else:
            #             t = 'xsd:string'
        return t

    def process_predicate(self, p: Optional[Union[URIRef, str]]) -> Tuple:
        """
        Process a predicate where the method checks if there is a mapping in Biolink Model.

        Parameters
        ----------
        p: Optional[Union[URIRef, str]]
            The predicate

        Returns
        -------
        Tuple
            A tuple that contains the Biolink CURIE (if available), the Biolink slot_uri CURIE (if available),
            the CURIE form of p, the reference of p

        """
        if p in self.cache:
            # already processed this predicate before; pull from cache
            element_uri = self.cache[p]["element_uri"]
            canonical_uri = self.cache[p]["canonical_uri"]
            predicate = self.cache[p]["predicate"]
            property_name = self.cache[p]["property_name"]
        else:
            # haven't seen this property before; map to element
            if self.prefix_manager.is_iri(p):
                predicate = self.prefix_manager.contract(str(p))
            else:
                predicate = None
            if self.prefix_manager.is_curie(p):
                property_name = self.prefix_manager.get_reference(p)
                predicate = p
            else:
                if predicate and self.prefix_manager.is_curie(predicate):
                    property_name = self.prefix_manager.get_reference(predicate)
                else:
                    property_name = p
                    predicate = f":{p}"
            element = self.get_biolink_element(p)
            canonical_uri = None
            if element:
                if isinstance(element, SlotDefinition):
                    # predicate corresponds to a biolink slot
                    if element.definition_uri:
                        element_uri = self.prefix_manager.contract(
                            element.definition_uri
                        )
                    else:
                        element_uri = (
                            f"biolink:{sentencecase_to_snakecase(element.name)}"
                        )
                    if element.slot_uri:
                        canonical_uri = element.slot_uri
                elif isinstance(element, ClassDefinition):
                    # this will happen only when the IRI is actually
                    # a reference to a class
                    element_uri = self.prefix_manager.contract(element.class_uri)
                else:
                    element_uri = f"biolink:{sentencecase_to_camelcase(element.name)}"
                if "biolink:Attribute" in get_biolink_ancestors(element.name):
                    element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
                if not predicate:
                    predicate = element_uri
            else:
                # no mapping to biolink model;
                # look at predicate mappings
                element_uri = None
                if p in self.reverse_predicate_mapping:
                    property_name = self.reverse_predicate_mapping[p]
                    predicate = f":{property_name}"
            self.cache[p] = {
                "element_uri": element_uri,
                "canonical_uri": canonical_uri,
                "predicate": predicate,
                "property_name": property_name,
            }
        return element_uri, canonical_uri, predicate, property_name

    def get_biolink_element(self, predicate: Any) -> Optional[Element]:
        """
        Returns a Biolink Model element for a given predicate.

        Parameters
        ----------
        predicate: Any
            The CURIE of a predicate

        Returns
        -------
        Optional[Element]
            The corresponding Biolink Model element

        """
        toolkit = get_toolkit()
        if self.prefix_manager.is_iri(predicate):
            predicate_curie = self.prefix_manager.contract(predicate)
        else:
            predicate_curie = predicate
        if self.prefix_manager.is_curie(predicate_curie):
            reference = self.prefix_manager.get_reference(predicate_curie)
        else:
            reference = predicate_curie
        element = toolkit.get_element(reference)
        if not element:
            try:
                mapping = toolkit.get_element_by_mapping(predicate)
                if mapping:
                    element = toolkit.get_element(mapping)
            except ValueError as e:
                log.error(e)
        return element

    def reify(self, u: str, v: str, data: Dict) -> Dict:
        """
        Create a node representation of an edge.

        Parameters
        ----------
        u: str
            Subject
        v: str
            Object
        k: str
            Edge key
        data: Dict
            Edge data

        Returns
        -------
        Dict
            The reified node

        """
        s = self.uriref(u)
        p = self.uriref(data["predicate"])
        o = self.uriref(v)

        if "id" in data:
            node_id = self.uriref(data["id"])
        else:
            # generate a UUID for the reified node
            node_id = self.uriref(generate_uuid())
        reified_node = data.copy()
        if "category" in reified_node:
            del reified_node["category"]
        reified_node["id"] = node_id
        reified_node["type"] = "biolink:Association"
        reified_node["subject"] = s
        reified_node["predicate"] = p
        reified_node["object"] = o
        return reified_node

    def finalize(self) -> None:
        """
        Perform any operations after writing the file.
        """
        self.FH.close()



"""
KGX Source for Simple Standard for Sharing Ontology Mappings ("SSSOM")
"""
import gzip
import re

import pandas as pd
from typing import Optional, Generator, Any, Dict, Tuple

import yaml

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.source import Source
from kgx.utils.kgx_utils import (
    validate_node,
    sanitize_import,
    validate_edge,
    generate_uuid,
    generate_edge_key,
)
from kgx.utils.rdf_utils import process_predicate

log = get_logger()

SSSOM_NODE_PROPERTY_MAPPING = {
    "subject_id": "id",
    "subject_category": "category",
    "object_id": "id",
    "object_category": "category",
}


class SssomSource(Source):
    """
    SssomSource is responsible for reading data as records
    from an SSSOM file.
    """

    def __init__(self):
        super().__init__()
        self.predicate_mapping = {}

    def set_prefix_map(self, m: Dict) -> None:
        """
        Add or override default prefix to IRI map.

        Parameters
        ----------
        m: Dict
            Prefix to IRI map

        """
        self.prefix_manager.set_prefix_map(m)

    def set_reverse_prefix_map(self, m: Dict) -> None:
        """
        Add or override default IRI to prefix map.

        Parameters
        ----------
        m: Dict
            IRI to prefix map

        """
        self.prefix_manager.set_reverse_prefix_map(m)

    def parse(
        self,
        filename: str,
        format: str,
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        Parse a SSSOM TSV

        Parameters
        ----------
        filename: str
            File to read from
        format: str
            The input file format (``tsv``, by default)
        compression: Optional[str]
            The compression (``gz``)
        kwargs: Dict
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        if "delimiter" not in kwargs:
            kwargs["delimiter"] = "\t"
        self.parse_header(filename, compression)

        # SSSOM 'mapping provider' may override the default 'knowledge_source' setting?
        if "mapping_provider" in self.graph_metadata:
            kwargs["knowledge_source"] = self.graph_metadata["mapping_provider"]

        self.set_provenance_map(kwargs)

        if compression:
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename)
        file_iter = pd.read_csv(
            FH,
            comment="#",
            dtype=str,
            chunksize=10000,
            low_memory=False,
            keep_default_na=False,
            **kwargs,
        )
        for chunk in file_iter:
            yield from self.load_edges(chunk)

    def parse_header(self, filename: str, compression: Optional[str] = None) -> None:
        """
        Parse metadata from SSSOM headers.

        Parameters
        ----------
        filename: str
            Filename to parse
        compression: Optional[str]
            Compression type

        """
        yamlstr = ""
        if compression:
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename)
        for line in FH:
            if line.startswith("#"):
                yamlstr += re.sub("^#", "", line)
            else:
                break
        if yamlstr:
            metadata = yaml.safe_load(yamlstr)
            log.info(f"Metadata: {metadata}")
            if "curie_map" in metadata:
                self.prefix_manager.update_prefix_map(metadata["curie_map"])
            for k, v in metadata.items():
                self.graph_metadata[k] = v

    def load_node(self, node: Dict) -> Tuple[str, Dict]:
        """
        Load a node into an instance of BaseGraph

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        Optional[Tuple[str, Dict]]
            A tuple that contains node id and node data

        """
        node = validate_node(node)
        node_data = sanitize_import(node.copy())
        if "id" in node_data:
            n = node_data["id"]

            self.set_node_provenance(node_data)

            self.node_properties.update(list(node_data.keys()))
            return n, node_data
        else:
            log.info("Ignoring node with no 'id': {}".format(node))

    def load_edges(self, df: pd.DataFrame) -> Generator:
        """
        Load edges from pandas.DataFrame into an instance of BaseGraph

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing records that represent edges

        Returns
        -------
        Generator
            A generator for edge records

        """
        for obj in df.to_dict("records"):
            yield from self.load_edge(obj)

    def load_edge(self, edge: Dict) -> Generator:
        """
        Load an edge into an instance of BaseGraph

        Parameters
        ----------
        edge : Dict
            An edge

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        (element_uri, canonical_uri, predicate, property_name) = process_predicate(
            self.prefix_manager, edge["predicate_id"], self.predicate_mapping
        )
        if element_uri:
            edge_predicate = element_uri
        elif predicate:
            edge_predicate = predicate
        else:
            edge_predicate = property_name
        if canonical_uri:
            edge_predicate = element_uri
        data = {
            "subject": edge["subject_id"],
            "predicate": edge_predicate,
            "object": edge["object_id"],
        }
        del edge["predicate_id"]
        data = validate_edge(data)
        subject_node = {}
        object_node = {}
        for k, v in edge.items():
            if k in SSSOM_NODE_PROPERTY_MAPPING:
                if k.startswith("subject"):
                    mapped_k = SSSOM_NODE_PROPERTY_MAPPING[k]
                    if mapped_k == "category" and not PrefixManager.is_curie(v):
                        v = f"biolink:OntologyClass"
                    subject_node[mapped_k] = v
                elif k.startswith("object"):
                    mapped_k = SSSOM_NODE_PROPERTY_MAPPING[k]
                    if mapped_k == "category" and not PrefixManager.is_curie(v):
                        v = f"biolink:OntologyClass"
                    object_node[mapped_k] = v
                else:
                    log.info(f"Ignoring {k} {v}")
            else:
                data[k] = v

        objs = [self.load_node(subject_node), self.load_node(object_node)]

        for k, v in self.graph_metadata.items():
            if k not in {"curie_map"}:
                data[k] = v

        edge_data = sanitize_import(data.copy())
        if "subject" in edge_data and "object" in edge_data:
            if "id" not in edge_data:
                edge_data["id"] = generate_uuid()
            s = edge_data["subject"]
            o = edge_data["object"]

            self.set_edge_provenance(edge_data)

            key = generate_edge_key(s, edge_data["predicate"], o)
            self.edge_properties.update(list(edge_data.keys()))
            objs.append((s, o, key, edge_data))
        else:
            log.info(
                "Ignoring edge with either a missing 'subject' or 'object': {}".format(
                    edge_data
                )
            )

        for o in objs:
            yield o



import gzip
from typing import Optional, Generator, Any

import ijson
from itertools import chain

from kgx.source.tsv_source import TsvSource


class JsonSource(TsvSource):
    """
    JsonSource is responsible for reading data as records
    from a JSON.
    """

    def __init__(self):
        super().__init__()
        self.compression = None

    def parse(
        self,
        filename: str,
        format: str = "json",
        compression: Optional[str] = None,
        **kwargs: Any
    ) -> Generator:
        """
        This method reads from a JSON and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``json``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records read from the file

        """

        self.set_provenance_map(kwargs)

        self.compression = compression
        n = self.read_nodes(filename)
        e = self.read_edges(filename)
        yield from chain(n, e)

    def read_nodes(self, filename: str) -> Generator:
        """
        Read node records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from

        Returns
        -------
        Generator
            A generator for node records

        """
        if self.compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for n in ijson.items(FH, "nodes.item"):
            yield self.read_node(n)

    def read_edges(self, filename: str) -> Generator:
        """
        Read edge records from a JSON.

        Parameters
        ----------
        filename: str
            The filename to read from

        Returns
        -------
        Generator
            A generator for edge records

        """
        if self.compression == "gz":
            FH = gzip.open(filename, "rb")
        else:
            FH = open(filename, "rb")
        for e in ijson.items(FH, "edges.item"):
            yield self.read_edge(e)



import copy
from typing import List

from kgx.config import get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.utils.kgx_utils import prepare_data_dict


log = get_logger()


def merge_all_graphs(graphs: List[BaseGraph], preserve: bool = True) -> BaseGraph:
    """
    Merge one or more graphs.

    .. note::
        This method will first pick the largest graph in ``graphs`` and use that
        as the target to merge the remaining graphs. This is to reduce the memory
        footprint for this operation. The criteria for largest graph is the graph
        with the largest number of edges.

        The caveat is that the merge operation has a side effect where the largest
        graph is altered.

        If you would like to ensure that all incoming graphs remain as-is, then
        look at ``merge_graphs``.

    The outcome of the merge on node and edge properties depend on the ``preserve`` parameter.
    If preserve is ``True`` then,
    - core properties will not be overwritten
    - other properties will be concatenated to a list

    If preserve is ``False`` then,
    - core properties will not be overwritten
    - other properties will be replaced

    Parameters
    ----------
    graphs: List[kgx.graph.base_graph.BaseGraph]
        A list of instances of BaseGraph to merge
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The merged graph

    """
    graph_size = [len(x.edges()) for x in graphs]
    largest = graphs.pop(graph_size.index(max(graph_size)))
    log.debug(
        f"Largest graph {largest.name} has {len(largest.nodes())} nodes and {len(largest.edges())} edges"
    )
    merged_graph = merge_graphs(largest, graphs, preserve)
    return merged_graph


def merge_graphs(
    graph: BaseGraph, graphs: List[BaseGraph], preserve: bool = True
) -> BaseGraph:
    """
    Merge all graphs in ``graphs`` to ``graph``.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        An instance of BaseGraph
    graphs: List[kgx.graph.base_graph.BaseGraph]
        A list of instances of BaseGraph to merge
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The merged graph

    """
    for g in graphs:
        node_merge_count = add_all_nodes(graph, g, preserve)
        log.info(
            f"Number of nodes merged between {graph.name} and {g.name}: {node_merge_count}"
        )
        edge_merge_count = add_all_edges(graph, g, preserve)
        log.info(
            f"Number of edges merged between {graph.name} and {g.name}: {edge_merge_count}"
        )
    return graph


def add_all_nodes(g1: BaseGraph, g2: BaseGraph, preserve: bool = True) -> int:
    """
    Add all nodes from source graph (``g2``) to target graph (``g1``).

    Parameters
    ----------
    g1: kgx.graph.base_graph.BaseGraph
        Target graph
    g2: kgx.graph.base_graph.BaseGraph
        Source graph
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    int
        Number of nodes merged during this operation

    """
    log.info(f"Adding {g2.number_of_nodes()} nodes from {g2.name} to {g1.name}")
    merge_count = 0
    for n, data in g2.nodes(data=True):
        if n in g1.nodes():
            merge_node(g1, n, data, preserve)
            merge_count += 1
        else:
            g1.add_node(n, **data)
    return merge_count


def merge_node(g: BaseGraph, n: str, data: dict, preserve: bool = True) -> dict:
    """
    Merge node ``n`` into graph ``g``.

    Parameters
    ----------
    g: kgx.graph.base_graph.BaseGraph
        The target graph
    n: str
        Node id
    data: dict
        Node properties
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    dict
        The merged node

    """
    existing_node = g.nodes()[n]
    new_data = prepare_data_dict(
        copy.deepcopy(existing_node), copy.deepcopy(data), preserve
    )
    g.add_node(n, **new_data)
    return existing_node


def add_all_edges(g1: BaseGraph, g2: BaseGraph, preserve: bool = True) -> int:
    """
    Add all edges from source graph (``g2``) to target graph (``g1``).

    Parameters
    ----------
    g1: kgx.graph.base_graph.BaseGraph
        Target graph
    g2: kgx.graph.base_graph.BaseGraph
        Source graph
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    int
        Number of edges merged during this operation

    """
    log.info(f"Adding {g2.number_of_edges()} edges from {g2} to {g1}")
    merge_count = 0
    for u, v, key, data in g2.edges(keys=True, data=True):
        if g1.has_edge(u, v, key):
            merge_edge(g1, u, v, key, data, preserve)
            merge_count += 1
        else:
            g1.add_edge(u, v, edge_key=key, **data)
    return merge_count


def merge_edge(
    g: BaseGraph, u: str, v: str, key: str, data: dict, preserve: bool = True
) -> dict:
    """
    Merge edge ``u`` -> ``v`` into graph ``g``.

    Parameters
    ----------
    g: kgx.graph.base_graph.BaseGraph
        The target graph
    u: str
        Subject node id
    v: str
        Object node id
    key: str
        Edge key
    data: dict
        Node properties
    preserve: bool
        Whether or not to preserve conflicting properties

    Returns
    -------
    dict
        The merged edge

    """
    existing_edge = g.get_edge(u, v, key)
    new_data = prepare_data_dict(
        copy.deepcopy(existing_edge), copy.deepcopy(data), preserve
    )
    g.add_edge(u, v, edge_key=key, **new_data)
    return existing_edge



import codecs
from typing import Generator

from rdflib.plugins.parsers.ntriples import NTriplesParser, ParseError
from rdflib.plugins.parsers.ntriples import r_wspace, r_wspaces, r_tail


class CustomNTriplesParser(NTriplesParser):
    """
    This class is an extension to ``rdflib.plugins.parsers.ntriples.NTriplesParser``
    that parses N-Triples and yields triples.
    """

    def parse(self, filename: str) -> Generator:
        """
        Parses an N-Triples file and yields triples.

        Parameters
        ----------
        filename: str
            The filename to parse

        Returns
        -------
        Generator
            A generator for triples

        """
        if not hasattr(filename, "read"):
            raise ParseError("Item to parse must be a file-like object.")

        # since N-Triples 1.1 files can and should be utf-8 encoded
        f = codecs.getreader("utf-8")(filename)

        self.file = f
        self.buffer = ""
        while True:
            self.line = self.readline()
            if self.line is None:
                break
            if self.line == "":
                break
            try:
                yield from self.parseline()
            except ParseError:
                raise ParseError("Invalid line: %r" % self.line)

    def parseline(self) -> Generator:
        """
        Parse each line and yield triples.

        Parameters
        ----------
        Generator
            A generator

        """
        print(self.line)
        self.eat(r_wspace)
        if self.line or not self.line.startswith("#"):
            subject = self.subject()
            self.eat(r_wspaces)

            predicate = self.predicate()
            self.eat(r_wspaces)

            object = self.object()
            self.eat(r_tail)

            if self.line:
                raise ParseError("Trailing garbage")
            return self.sink.triple(subject, predicate, object)



from setuptools import setup, find_packages

NAME = 'kgx'
DESCRIPTION = 'A Python library and set of command line utilities for exchanging Knowledge Graphs (KGs) that conform to or are aligned to the Biolink Model.'
URL = 'https://github.com/NCATS-Tangerine/kgx'
AUTHOR = 'Deepak Unni, Sierra Moxon, Richard Bruskiewich'
EMAIL = 'deepak.unni3@gmail.com, smoxon@lbl.gov, richard.bruskiewich@delphinai.com'
REQUIRES_PYTHON = '>=3.7.0'
VERSION = '1.5.1'
LICENSE = 'BSD'

with open("requirements.txt", "r") as FH:
    REQUIREMENTS = FH.readlines()

EXTRAS = {}

setup(
    name=NAME,
    author=AUTHOR,
    version=VERSION,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    description=DESCRIPTION,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license=LICENSE,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_data={'kgx': ["config.yml"]},
    keywords='knowledge-graph Neo4j RDF NCATS NCATS-Translator Biolink-Model',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3'
    ],
    install_requires=[r for r in REQUIREMENTS if not r.startswith("#")],
    extras_require=EXTRAS,
    include_package_data=True,
    entry_points={
        'console_scripts': ['kgx=kgx.cli:cli']
    }
)



import gzip
import os
from typing import Optional, Dict, Any

import jsonlines

from kgx.sink.sink import Sink


class JsonlSink(Sink):
    """
    JsonlSink is responsible for writing data as records
    to JSON lines.

    Parameters
    ----------
    filename: str
        The filename to write to
    format: str
        The file format (``jsonl``)
    compression: Optional[str]
        The compression type (``gz``)
    kwargs: Any
        Any additional arguments

    """

    def __init__(
        self,
        filename: str,
        format: str = "jsonl",
        compression: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()
        dirname = os.path.abspath(os.path.dirname(filename))
        basename = os.path.basename(filename)
        nodes_filename = os.path.join(
            dirname if dirname else "", f"{basename}_nodes.{format}"
        )
        edges_filename = os.path.join(
            dirname if dirname else "", f"{basename}_edges.{format}"
        )
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        if compression == "gz":
            nodes_filename += f".{compression}"
            edges_filename += f".{compression}"
            NFH = gzip.open(nodes_filename, "wb")
            self.NFH = jsonlines.Writer(NFH)
            EFH = gzip.open(edges_filename, "wb")
            self.EFH = jsonlines.Writer(EFH)
        else:
            self.NFH = jsonlines.open(nodes_filename, "w")
            self.EFH = jsonlines.open(edges_filename, "w")

    def write_node(self, record: Dict) -> None:
        """
        Write a node record to JSON.

        Parameters
        ----------
        record: Dict
            A node record

        """
        self.NFH.write(record)

    def write_edge(self, record: Dict) -> None:
        """
        Write an edge record to JSON.

        Parameters
        ----------
        record: Dict
            A node record

        """
        self.EFH.write(record)

    def finalize(self) -> None:
        """
        Perform any operations after writing the file.
        """
        self.NFH.close()
        self.EFH.close()
